from __future__ import print_function

import sys
from statistics import mean

from svim.SVSignature import SignatureDeletion, SignatureInsertion, SignatureInversion, SignatureTranslocation, SignatureDuplicationTandem, SignatureInsertionFrom
from svim.SVIM_clustering import consolidate_clusters_bilocal, clusters_from_partitions


def is_similar(chr1, start1, end1, chr2, start2, end2):
    if chr1 == chr2 and abs(start1 - start2) < 20 and abs(end1 - end2) < 20:
        return True
    else:
        return False


def analyze_read_segments(primary, supplementaries, bam, options):
    if len(supplementaries) > 3:
        return []

    read_name = primary.query_name
    alignments = [primary] + supplementaries
    alignment_list = []
    for alignment in alignments:
        #correct query coordinates for reversely mapped reads
        if alignment.is_reverse:
            q_start = alignment.infer_read_length() - alignment.query_alignment_end
            q_end = alignment.infer_read_length() - alignment.query_alignment_start
        else:
            q_start = alignment.query_alignment_start
            q_end = alignment.query_alignment_end

        new_alignment_dict = {  'q_start': q_start, 
                                'q_end': q_end, 
                                'ref_id': alignment.reference_id, 
                                'ref_start': alignment.reference_start, 
                                'ref_end': alignment.reference_end,
                                'is_reverse': alignment.is_reverse  }
        alignment_list.append(new_alignment_dict)

    sorted_alignment_list = sorted(alignment_list, key=lambda aln: (aln['q_start'], aln['q_end']))
    #inferred_read_length = alignments[0].infer_read_length()

    sv_signatures = []
    tandem_duplications = []
    translocations = []

    for index in range(len(sorted_alignment_list) - 1):
        alignment_current = sorted_alignment_list[index]
        alignment_next = sorted_alignment_list[index + 1]

        distance_on_read = alignment_next['q_start'] - alignment_current['q_end']

        #Same chromosome
        if alignment_current['ref_id'] == alignment_next['ref_id']:
            ref_chr = bam.getrname(alignment_current['ref_id'])
            #Same orientation
            if alignment_current['is_reverse'] == alignment_next['is_reverse']:
                #Compute distance on reference depending on orientation
                if alignment_current['is_reverse']:
                    distance_on_reference = alignment_current['ref_start'] - alignment_next['ref_end']
                else:
                    distance_on_reference = alignment_next['ref_start'] - alignment_current['ref_end']
                #No overlap on read
                if distance_on_read >= -options.segment_overlap_tolerance:
                    #No overlap on reference
                    if distance_on_reference >= -options.segment_overlap_tolerance:
                        deviation = distance_on_read - distance_on_reference
                        #INS candidate
                        if deviation >= options.min_sv_size:
                            #No gap on reference
                            if distance_on_reference <= options.segment_gap_tolerance:
                                if not alignment_current['is_reverse']:
                                    sv_signatures.append(SignatureInsertion(ref_chr, alignment_current['ref_end'], alignment_current['ref_end'] + deviation, "suppl", read_name))
                                else:
                                    sv_signatures.append(SignatureInsertion(ref_chr, alignment_current['ref_start'], alignment_current['ref_start'] + deviation, "suppl", read_name))
                        #DEL candidate
                        elif -options.max_sv_size <= deviation <= -options.min_sv_size:
                            #No gap on read
                            if distance_on_read <= options.segment_gap_tolerance:
                                if not alignment_current['is_reverse']:
                                    sv_signatures.append(SignatureDeletion(ref_chr, alignment_current['ref_end'], alignment_current['ref_end'] - deviation, "suppl", read_name))
                                else:
                                    sv_signatures.append(SignatureDeletion(ref_chr, alignment_next['ref_end'], alignment_next['ref_end'] - deviation, "suppl", read_name))
                        #Either very large DEL or TRANS
                        elif deviation < -options.max_sv_size:
                            #No gap on read
                            if distance_on_read <= options.segment_gap_tolerance:
                                if not alignment_current['is_reverse']:
                                    sv_signatures.append(SignatureTranslocation(ref_chr, alignment_current['ref_end'] - 1, 'fwd', ref_chr, alignment_next['ref_start'], 'fwd', "suppl", read_name))
                                    translocations.append(('fwd', 'fwd', ref_chr, alignment_current['ref_end'] - 1, ref_chr, alignment_next['ref_start']))
                                else:
                                    sv_signatures.append(SignatureTranslocation(ref_chr, alignment_current['ref_start'], 'rev', ref_chr, alignment_next['ref_end'] - 1, 'rev', "suppl", read_name))
                                    translocations.append(('rev', 'rev', ref_chr, alignment_current['ref_start'], ref_chr, alignment_next['ref_end'] - 1))
                    #overlap on reference
                    else:
                        #Tandem Duplication
                        if distance_on_reference < -options.min_sv_size:
                            if not alignment_current['is_reverse']:
                                #Tandem Duplication
                                if alignment_next['ref_end'] > alignment_current['ref_start']:
                                    tandem_duplications.append((ref_chr, alignment_next['ref_start'], alignment_current['ref_end']))
                                #Either very large TANDEM or TRANS
                                else:
                                    sv_signatures.append(SignatureTranslocation(ref_chr, alignment_current['ref_end'] - 1, 'fwd', ref_chr, alignment_next['ref_start'], 'fwd', "suppl", read_name))
                                    translocations.append(('fwd', 'fwd', ref_chr, alignment_current['ref_end'] - 1, ref_chr, alignment_next['ref_start']))
                            else:
                                #Tandem Duplication
                                if alignment_next['ref_start'] < alignment_current['ref_end']:
                                    tandem_duplications.append((ref_chr, alignment_current['ref_start'], alignment_next['ref_end']))
                                #Either very large TANDEM or TRANS
                                else:
                                    sv_signatures.append(SignatureTranslocation(ref_chr, alignment_current['ref_start'], 'rev', ref_chr, alignment_next['ref_end'] - 1, 'rev', "suppl", read_name))
                                    translocations.append(('rev', 'rev', ref_chr, alignment_current['ref_start'], ref_chr, alignment_next['ref_end'] - 1))
            #Different orientations
            else:
                #Normal to reverse
                if not alignment_current['is_reverse'] and alignment_next['is_reverse']:
                    if -options.segment_overlap_tolerance <= distance_on_read <= options.segment_gap_tolerance:
                        if alignment_next['ref_start'] - alignment_current['ref_end'] >= -options.segment_overlap_tolerance: # Case 1
                            #INV candidate
                            if alignment_next['ref_end'] - alignment_current['ref_end'] <= options.max_sv_size:
                                sv_signatures.append(SignatureInversion(ref_chr, alignment_current['ref_end'], alignment_next['ref_end'], "suppl", read_name, "left_fwd"))
                                #transitions.append(('inversion', 'left_fwd', ref_chr, alignment_current['ref_end'], alignment_next['ref_end']))
                            #Either very large INV or TRANS
                            else:
                                sv_signatures.append(SignatureTranslocation(ref_chr, alignment_current['ref_end'] - 1, 'fwd', ref_chr, alignment_next['ref_end'] - 1, 'rev', "suppl", read_name))
                                translocations.append(('fwd', 'rev', ref_chr, alignment_current['ref_end'] - 1, ref_chr, alignment_next['ref_end'] - 1))
                        elif alignment_current['ref_start'] - alignment_next['ref_end'] >= -options.segment_overlap_tolerance: # Case 3
                            #INV candidate
                            if alignment_current['ref_end'] - alignment_next['ref_end'] <= options.max_sv_size:
                                sv_signatures.append(SignatureInversion(ref_chr, alignment_next['ref_end'], alignment_current['ref_end'], "suppl", read_name, "left_rev"))
                                #transitions.append(('inversion', 'left_rev', ref_chr, alignment_next['ref_end'], alignment_current['ref_end']))
                            #Either very large INV or TRANS
                            else:
                                sv_signatures.append(SignatureTranslocation(ref_chr, alignment_current['ref_end'] - 1, 'fwd', ref_chr, alignment_next['ref_end'] - 1, 'rev', "suppl", read_name))
                                translocations.append(('fwd', 'rev', ref_chr, alignment_current['ref_end'] - 1, ref_chr, alignment_next['ref_end'] - 1))
                    else:
                        pass
                        #print("Overlapping read segments in read", read_name)
                #Reverse to normal
                if alignment_current['is_reverse'] and not alignment_next['is_reverse']:
                    if -options.segment_overlap_tolerance <= distance_on_read <= options.segment_gap_tolerance:
                        if alignment_next['ref_start'] - alignment_current['ref_end'] >= -options.segment_overlap_tolerance: # Case 2
                            #INV candidate
                            if alignment_next['ref_start'] - alignment_current['ref_start'] <= options.max_sv_size:
                                sv_signatures.append(SignatureInversion(ref_chr, alignment_current['ref_start'], alignment_next['ref_start'], "suppl", read_name, "right_fwd"))
                                #transitions.append(('inversion', 'right_fwd', ref_chr, alignment_current['ref_start'], alignment_next['ref_start']))
                            #Either very large INV or TRANS
                            else:
                                sv_signatures.append(SignatureTranslocation(ref_chr, alignment_current['ref_start'], 'rev', ref_chr, alignment_next['ref_start'], 'fwd', "suppl", read_name))
                                translocations.append(('rev', 'fwd', ref_chr, alignment_current['ref_start'], ref_chr, alignment_next['ref_start']))
                        elif alignment_current['ref_start'] - alignment_next['ref_end'] >= -options.segment_overlap_tolerance: # Case 4
                            #INV candidate
                            if alignment_current['ref_start'] - alignment_next['ref_start'] <= options.max_sv_size:
                                sv_signatures.append(SignatureInversion(ref_chr, alignment_next['ref_start'], alignment_current['ref_start'], "suppl", read_name, "right_rev"))
                                #transitions.append(('inversion', 'right_rev', ref_chr, alignment_next['ref_start'], alignment_current['ref_start']))
                            #Either very large INV or TRANS
                            else:
                                sv_signatures.append(SignatureTranslocation(ref_chr, alignment_current['ref_start'], 'rev', ref_chr, alignment_next['ref_start'], 'fwd', "suppl", read_name))
                                translocations.append(('rev', 'fwd', ref_chr, alignment_current['ref_start'], ref_chr, alignment_next['ref_start']))
                    else:
                        pass
                        #print("Overlapping read segments in read", read_name)
        #Different chromosomes
        else:
            ref_chr_current = bam.getrname(alignment_current['ref_id'])
            ref_chr_next = bam.getrname(alignment_next['ref_id'])
            #Same orientation
            if alignment_current['is_reverse'] == alignment_next['is_reverse']:
                #No overlap on read
                if distance_on_read >= -options.segment_overlap_tolerance:
                    #No gap on read
                    if distance_on_read <= options.segment_gap_tolerance:
                        if not alignment_current['is_reverse']:
                            sv_signatures.append(SignatureTranslocation(ref_chr_current, alignment_current['ref_end'] - 1, 'fwd', ref_chr_next, alignment_next['ref_start'], 'fwd', "suppl", read_name))
                            translocations.append(('fwd', 'fwd', ref_chr_current, alignment_current['ref_end'] - 1, ref_chr_next, alignment_next['ref_start']))
                        else:
                            sv_signatures.append(SignatureTranslocation(ref_chr_current, alignment_current['ref_start'], 'rev', ref_chr_next, alignment_next['ref_end'] - 1, 'rev', "suppl", read_name))
                            translocations.append(('rev', 'rev', ref_chr_current, alignment_current['ref_start'], ref_chr_next, alignment_next['ref_end'] - 1))
                #Overlap on read
                else:
                    pass
                    #print("Overlapping read segments in read", read_name)
            #Different orientation
            else:
                #INV + TRANS
                pass

    #Handle tandem duplications
    current_chromosome = None
    current_starts = []
    current_ends = []
    current_copy_number = 0
    for tandem_duplication in tandem_duplications:
        if current_chromosome == None:
            current_chromosome = tandem_duplication[0]
            current_starts.append(tandem_duplication[1])
            current_ends.append(tandem_duplication[2])
            current_copy_number = 1
        else:
            if is_similar(current_chromosome, mean(current_starts), mean(current_ends), tandem_duplication[0], tandem_duplication[1], tandem_duplication[2]):
                current_starts.append(tandem_duplication[1])
                current_ends.append(tandem_duplication[2])
                current_copy_number += 1
            else:
                sv_signatures.append(SignatureDuplicationTandem(current_chromosome, int(mean(current_starts)), int(mean(current_ends)), current_copy_number, "suppl", read_name))
                current_chromosome = tandem_duplication[0]
                current_starts.append(tandem_duplication[1])
                current_ends.append(tandem_duplication[2])
                current_copy_number = 1
    if current_chromosome != None:
        sv_signatures.append(SignatureDuplicationTandem(current_chromosome, int(mean(current_starts)), int(mean(current_ends)), current_copy_number, "suppl", read_name))

    #Handle insertions_from
    for this_index in range(len(translocations)):
        this_dir1 = translocations[this_index][0]
        this_dir2 = translocations[this_index][1]
        this_chr1 = translocations[this_index][2]
        this_pos1 = translocations[this_index][3]
        this_chr2 = translocations[this_index][4]
        this_pos2 = translocations[this_index][5]

        for before_dir1, before_dir2, before_chr1, before_pos1, before_chr2, before_pos2 in translocations[:this_index]:
            #Same direction at destination and origin
            if before_dir1 == this_dir2 and before_dir2 == this_dir1:
                #Same position at destination
                if is_similar(before_chr1, before_pos1, 0, this_chr2, this_pos2, 0):
                    #Same chromosome for origin
                    if before_chr2 == this_chr1:
                        #INS_DUP candidate
                        if before_dir2 == before_dir1:
                            if before_dir1 == 'fwd':
                                if options.min_sv_size <= this_pos1 - before_pos2 + 1 <= options.max_sv_size:
                                    sv_signatures.append(SignatureInsertionFrom(before_chr2, before_pos2, this_pos1 + 1, before_chr1, int(mean([before_pos1 + 1, this_pos2])), "suppl", read_name))
                            elif before_dir1 == 'rev':
                                if options.min_sv_size <= before_pos2 - this_pos1 <= options.max_sv_size:
                                    sv_signatures.append(SignatureInsertionFrom(before_chr2, this_pos1, before_pos2 + 1, before_chr1, int(mean([before_pos1, this_pos2 + 1])), "suppl", read_name))
                        #INV_INS_DUP candidate
                        else:
                            pass

    return sv_signatures



from __future__ import print_function

import sys
from bisect import bisect_left
from collections import defaultdict
from math import pow, sqrt

from svim.SVSignature import SignatureTranslocation, SignatureInsertionFrom, SignatureClusterBiLocal
from svim.SVCandidate import CandidateDuplicationInterspersed
from svim.SVIM_clustering import span_position_distance

def flag_cutpaste_candidates(insertion_from_signature_clusters, deletion_signature_clusters, options):
    """Flag duplication signature clusters if they overlap a deletion"""
    int_duplication_candidates = []
    for ins_cluster in insertion_from_signature_clusters:
        # Compute distances of every deletion cluster to the current insertion/duplication
        distances = [(del_index, span_position_distance((del_cluster.get_source()[1], del_cluster.get_source()[2], options.distance_normalizer), \
                                                        (ins_cluster.get_source()[1], ins_cluster.get_source()[2], options.distance_normalizer))) \
                                                        for del_index, del_cluster in enumerate(deletion_signature_clusters)]
        closest_deletion_index, closest_deletion = sorted(distances, key=lambda obj: obj[1])[0]
        source_contig, source_start, source_end = ins_cluster.get_source()
        dest_contig, dest_start, dest_end = ins_cluster.get_destination()
        # If close deletion cluster found
        if closest_deletion <= options.del_ins_dup_max_distance:
            #Potential cut&paste insertion
            int_duplication_candidates.append(CandidateDuplicationInterspersed(source_contig, source_start, source_end, dest_contig, dest_start, dest_end, ins_cluster.members, ins_cluster.score, ins_cluster.std_span, ins_cluster.std_pos, cutpaste=True))
        else:
            #Interspersed duplication
            int_duplication_candidates.append(CandidateDuplicationInterspersed(source_contig, source_start, source_end, dest_contig, dest_start, dest_end, ins_cluster.members, ins_cluster.score, ins_cluster.std_span, ins_cluster.std_pos, cutpaste=False))
    return int_duplication_candidates


def get_closest_index(input_list, input_number):
    """
    Assumes input_list is sorted. Returns index of closest value to input_number.

    If two numbers are equally close, return the index of the smallest number.
    """
    if len(input_list) < 1:
        return None
    pos = bisect_left(input_list, input_number)
    if pos == 0:
        return 0
    if pos == len(input_list):
        return len(input_list) - 1
    before = input_list[pos - 1]
    after = input_list[pos]
    if after - input_number < input_number - before:
       return pos
    else:
       return pos - 1


def distance_positions(position1, position2):
    return float("inf") if position1[0] != position2[0] else abs(position1[1] - position2[1])


def cluster_positions_simple(positions, max_delta):
    """Form partitions of (contig, position) pairs using simple distance."""
    sorted_positions = sorted(positions, key=lambda position: (position[0], position[1]))
    partitions = []
    current_partition = []
    for position in sorted_positions:
        if len(current_partition) < 1:
            current_partition.append(position)
            continue
        if distance_positions(current_partition[0], position) > max_delta:
            partitions.append(current_partition[:])
            while len(current_partition) > 0 and distance_positions(current_partition[0], position) > max_delta:
                current_partition.pop(0)
        current_partition.append(position)
    if len(current_partition) > 0:
        partitions.append(current_partition[:])
    return partitions


def calculate_score_deletion(main_score, translocation_distances, translocation_stds, translocation_deviation):
    """Calculate the score of a merged insertion detected from a deletion.
       Parameters: - main_score - score of the underlying main deletion
                   - translocation_distances - mean distance of the translocation clusters flanking the main deletion (left, right or both sides)
                   - translocation_stds - standard deviation of the translocation clusters flanking the main deletion (left, right or both sides)
                   - translocation_deviation - distance between left and right translocation destinations"""
    if len(translocation_distances) == 2:
        #scale translocation distances to [0, 1] range
        td0 = max(0, 100 - translocation_distances[0]) / 100
        td1 = max(0, 100 - translocation_distances[1]) / 100

        #scale translocation stds to [0, 1] range
        ts0 = max(0, 100 - translocation_stds[0]) / 100
        ts1 = max(0, 100 - translocation_stds[1]) / 100

        #scale translocation deviation to [0, 1] range
        tv = max(0, 100 - translocation_deviation) / 100

        #calculate final score as product of components
        product = (main_score / 100) * td0 * td1 * ts0 * ts1 * tv
        final_score = pow(product, 1/6) * 100
    if len(translocation_distances) == 1:
        #scale translocation distance to [0, 1] range
        td0 = max(0, 100 - translocation_distances[0]) / 100

        #scale translocation std to [0, 1] range
        ts0 = max(0, 100 - translocation_stds[0]) / 100

        #scale translocation deviation to [0, 1] range
        tv = max(0, 100 - translocation_deviation) / 100

        #calculate final score as product of components (half score)
        product = (main_score / 100) * td0 * td0 * ts0 * ts0 * tv
        final_score = pow(product, 1/6) * 50
    return final_score


def calculate_score_insertion(main_score, translocation_distances, translocation_stds, destination_stds):
    """Calculate the score of a merged insertion or duplication detected from an insertion.
       Parameters: - main_score - score of the underlying main insertion
                   - translocation_distances - mean distance of the translocation clusters flanking the main insertion (left)
                   - translocation_stds - standard deviation of the translocation clusters flanking the main insertion (left)
                   - destination_stds - standard deviations of the left and right translocation destinations"""
    #scale translocation distance to [0, 1] range
    td0 = max(0, 100 - translocation_distances[0]) / 100
    td1 = max(0, 100 - translocation_distances[1]) / 100

    #scale translocation std to [0, 1] range
    ts0 = max(0, 100 - translocation_stds[0]) / 100
    ts1 = max(0, 100 - translocation_stds[1]) / 100

    #scale destination stds to [0, 1] range
    ds0 = max(0, 100 - destination_stds[0]) / 100
    ds1 = max(0, 100 - destination_stds[1]) / 100

    #calculate final score as product of components
    product = (main_score / 100) *  td0 * td1 * ts0 * ts1 * ds0 * ds1
    final_score = pow(product, 1/7) * 100
    return final_score


def merge_translocations_at_insertions(translocation_partitions_fwdfwd_dict, translocation_partition_means_fwdfwd_dict, translocation_partition_stds_fwdfwd_dict, translocation_partitions_revrev_dict, translocation_partition_means_revrev_dict, translocation_partition_stds_revrev_dict, insertion_signature_clusters, options):
    inserted_regions_to_remove = []
    insertion_from_signature_clusters = []
    for insertion_index, ins_cluster in enumerate(insertion_signature_clusters):
        ins_contig, ins_start, ins_end = ins_cluster.get_source()
        try:
            closest_to_start_fwdfwd_index = get_closest_index(translocation_partition_means_fwdfwd_dict[ins_contig], ins_start)
            closest_to_start_fwdfwd_mean = translocation_partition_means_fwdfwd_dict[ins_contig][closest_to_start_fwdfwd_index]
            closest_to_start_revrev_index = get_closest_index(translocation_partition_means_revrev_dict[ins_contig], ins_start)
            closest_to_start_revrev_mean = translocation_partition_means_revrev_dict[ins_contig][closest_to_start_revrev_index]
        except KeyError:
            continue
        # if translocations found close to start of insertion
        if abs(closest_to_start_fwdfwd_mean - ins_start) <= options.trans_sv_max_distance and abs(closest_to_start_revrev_mean - ins_start) <= options.trans_sv_max_distance:
            destinations_from_start_fwdfwd = cluster_positions_simple([(signature.contig2, signature.pos2) for signature in translocation_partitions_fwdfwd_dict[ins_contig][closest_to_start_fwdfwd_index]], options.trans_destination_partition_max_distance)
            destinations_from_start_revrev = cluster_positions_simple([(signature.contig2, signature.pos2) for signature in translocation_partitions_revrev_dict[ins_contig][closest_to_start_revrev_index]], options.trans_destination_partition_max_distance)
            # if translocations point to only one destination each
            if len(destinations_from_start_fwdfwd) == 1 and len(destinations_from_start_revrev) == 1:
                destination_from_start_fwdfwd = (destinations_from_start_fwdfwd[0][0][0], int(round(sum([pos for contig, pos in destinations_from_start_fwdfwd[0]]) / len(destinations_from_start_fwdfwd[0]))))
                destination_from_start_fwdfwd_std = int(round(sqrt(sum([pow(abs(pos - destination_from_start_fwdfwd[1]), 2) for contig, pos in destinations_from_start_fwdfwd[0]]) / len(destinations_from_start_fwdfwd[0]))))
                destination_from_start_revrev = (destinations_from_start_revrev[0][0][0], int(round(sum([pos for contig, pos in destinations_from_start_revrev[0]]) / len(destinations_from_start_revrev[0]))))
                destination_from_start_revrev_std = int(round(sqrt(sum([pow(abs(pos - destination_from_start_revrev[1]), 2) for contig, pos in destinations_from_start_revrev[0]]) / len(destinations_from_start_revrev[0]))))
                # if the two destinations have the right distance
                distance = abs(destination_from_start_revrev[1] - destination_from_start_fwdfwd[1])
                if destination_from_start_revrev[0] == destination_from_start_fwdfwd[0] and 0.95 <= ((ins_end - ins_start + 1) / (distance + 1)) <= 1.1:
                    members = ins_cluster.members + translocation_partitions_fwdfwd_dict[ins_contig][closest_to_start_fwdfwd_index] + translocation_partitions_revrev_dict[ins_contig][closest_to_start_revrev_index]
                    score = calculate_score_insertion(ins_cluster.score, 
                                                      [abs(closest_to_start_fwdfwd_mean - ins_start), abs(closest_to_start_revrev_mean - ins_start)], 
                                                      [translocation_partition_stds_fwdfwd_dict[ins_contig][closest_to_start_fwdfwd_index], translocation_partition_stds_revrev_dict[ins_contig][closest_to_start_revrev_index]], 
                                                      [destination_from_start_fwdfwd_std, destination_from_start_revrev_std])
                    insertion_from_signature_clusters.append(SignatureClusterBiLocal(destination_from_start_revrev[0], min(destination_from_start_revrev[1], destination_from_start_fwdfwd[1]), max(destination_from_start_revrev[1], destination_from_start_fwdfwd[1]), ins_contig, ins_start, ins_start + distance, score, len(members), members, "ins_dup", ins_cluster.std_span, ins_cluster.std_pos))
                    inserted_regions_to_remove.append(insertion_index)

    return insertion_from_signature_clusters, inserted_regions_to_remove



import os
import logging
import re

from collections import defaultdict
from math import pow, sqrt

from svim.SVIM_clustering import form_partitions, partition_and_cluster_candidates
from svim.SVCandidate import CandidateInversion, CandidateDuplicationTandem, CandidateDeletion, CandidateNovelInsertion
from svim.SVIM_merging import flag_cutpaste_candidates, merge_translocations_at_insertions


def cluster_sv_candidates(int_duplication_candidates, options):
    """Cluster SVCandidates to remove redundancy"""

    final_int_duplication_candidates = partition_and_cluster_candidates(int_duplication_candidates, options, "interspersed duplication candidates")

    return final_int_duplication_candidates


def write_candidates(working_dir, candidates):
    int_duplication_candidates, inversion_candidates, tan_duplication_candidates, deletion_candidates, novel_insertion_candidates = candidates

    if not os.path.exists(working_dir + '/candidates'):
        os.mkdir(working_dir + '/candidates')
    deletion_candidate_output = open(working_dir + '/candidates/candidates_deletions.bed', 'w')
    inversion_candidate_output = open(working_dir + '/candidates/candidates_inversions.bed', 'w')
    tandem_duplication_candidate_source_output = open(working_dir + '/candidates/candidates_tan_duplications_source.bed', 'w')
    tandem_duplication_candidate_dest_output = open(working_dir + '/candidates/candidates_tan_duplications_dest.bed', 'w')
    interspersed_duplication_candidate_source_output = open(working_dir + '/candidates/candidates_int_duplications_source.bed', 'w')
    interspersed_duplication_candidate_dest_output = open(working_dir + '/candidates/candidates_int_duplications_dest.bed', 'w')
    novel_insertion_candidate_output = open(working_dir + '/candidates/candidates_novel_insertions.bed', 'w')

    for candidate in deletion_candidates:
        print(candidate.get_bed_entry(), file=deletion_candidate_output)
    for candidate in int_duplication_candidates:
        bed_entries = candidate.get_bed_entries()
        print(bed_entries[0], file=interspersed_duplication_candidate_source_output)
        print(bed_entries[1], file=interspersed_duplication_candidate_dest_output)
    for candidate in inversion_candidates:
        print(candidate.get_bed_entry(), file=inversion_candidate_output)
    for candidate in tan_duplication_candidates:
        bed_entries = candidate.get_bed_entries()
        print(bed_entries[0], file=tandem_duplication_candidate_source_output)
        print(bed_entries[1], file=tandem_duplication_candidate_dest_output)
    for candidate in novel_insertion_candidates:
        print(candidate.get_bed_entry(), file=novel_insertion_candidate_output)

    deletion_candidate_output.close()
    inversion_candidate_output.close()
    interspersed_duplication_candidate_source_output.close()
    interspersed_duplication_candidate_dest_output.close()
    tandem_duplication_candidate_source_output.close()
    tandem_duplication_candidate_dest_output.close()


def sorted_nicely(vcf_entries):
    """ Sort the given vcf entries (in the form ((contig, start, end), vcf_string)) in the way that humans expect.
        e.g. chr10 comes after chr2
        Algorithm adapted from https://blog.codinghorror.com/sorting-for-humans-natural-sort-order/"""
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ]
    tuple_key = lambda entry: ( alphanum_key(str(entry[0][0])), entry[0][1], entry[0][2] )
    return sorted(vcf_entries, key = tuple_key)


def write_final_vcf(working_dir, int_duplication_candidates, inversion_candidates, tandem_duplication_candidates, deletion_candidates, novel_insertion_candidates, version, contig_names, contig_lengths, sample):
    vcf_output = open(working_dir + '/final_results.vcf', 'w')

    # Write header lines
    print("##fileformat=VCFv4.2", file=vcf_output)
    print("##source=SVIMV{0}".format(version), file=vcf_output)
    #print("##reference={0}".format(genome), file=vcf_output)
    for contig_name, contig_length in zip(contig_names, contig_lengths):
        print("##contig=<ID={0},length={1}>".format(contig_name, contig_length), file=vcf_output)
    print("##ALT=<ID=DEL,Description=\"Deletion\">", file=vcf_output)
    print("##ALT=<ID=INV,Description=\"Inversion\">", file=vcf_output)
    print("##ALT=<ID=DUP,Description=\"Duplication\">", file=vcf_output)
    print("##ALT=<ID=DUP:TANDEM,Description=\"Tandem Duplication\">", file=vcf_output)
    print("##ALT=<ID=DUP:INT,Description=\"Interspersed Duplication\">", file=vcf_output)
    print("##ALT=<ID=INS,Description=\"Insertion\">", file=vcf_output)
    print("##ALT=<ID=INS:NOVEL,Description=\"Novel Insertion\">", file=vcf_output)
    print("##INFO=<ID=SVTYPE,Number=1,Type=String,Description=\"Type of structural variant\">", file=vcf_output)
    print("##INFO=<ID=CUTPASTE,Number=0,Type=Flag,Description=\"Genomic origin of interspersed duplication seems to be deleted\">", file=vcf_output)
    print("##INFO=<ID=END,Number=1,Type=Integer,Description=\"End position of the variant described in this record\">", file=vcf_output)
    print("##INFO=<ID=SVLEN,Number=1,Type=Integer,Description=\"Difference in length between REF and ALT alleles\">", file=vcf_output)
    print("##INFO=<ID=SUPPORT,Number=1,Type=Integer,Description=\"Number of reads supporting this variant\">", file=vcf_output)
    print("##INFO=<ID=STD_SPAN,Number=1,Type=Float,Description=\"Standard deviation in span of merged SV signatures\">", file=vcf_output)
    print("##INFO=<ID=STD_POS,Number=1,Type=Float,Description=\"Standard deviation in position of merged SV signatures\">", file=vcf_output)
    print("##FILTER=<ID=q5,Description=\"Score below 5\">", file=vcf_output)
    print("##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">", file=vcf_output)
    print("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t" + sample, file=vcf_output)

    vcf_entries = []
    for candidate in deletion_candidates:
        vcf_entries.append((candidate.get_source(), candidate.get_vcf_entry()))
    for candidate in inversion_candidates:
        vcf_entries.append((candidate.get_source(), candidate.get_vcf_entry()))
    for candidate in tandem_duplication_candidates:
        vcf_entries.append((candidate.get_destination(), candidate.get_vcf_entry()))
    for candidate in int_duplication_candidates:
        vcf_entries.append((candidate.get_destination(), candidate.get_vcf_entry()))
    for candidate in novel_insertion_candidates:
        vcf_entries.append((candidate.get_destination(), candidate.get_vcf_entry()))

    # Sort and write entries to VCF
    for source, entry in sorted_nicely(vcf_entries):
        print(entry, file=vcf_output)

    vcf_output.close()


def combine_clusters(signature_clusters, working_dir, options, version, contig_names, contig_lengths, sample):
    deletion_signature_clusters, insertion_signature_clusters, inversion_signature_clusters, tandem_duplication_signature_clusters, insertion_from_signature_clusters, completed_translocations = signature_clusters

    ###############################
    # Create inversion candidates #
    ###############################
    inversion_candidates = []
    for inv_cluster in inversion_signature_clusters:
        inversion_candidates.append(CandidateInversion(inv_cluster.contig, inv_cluster.start, inv_cluster.end, inv_cluster.members, inv_cluster.score, inv_cluster.std_span, inv_cluster.std_pos))

    ########################################
    # Create tandem duplication candidates #
    ########################################
    tan_dup_candidates = []
    for tan_dup_cluster in tandem_duplication_signature_clusters:
        source_contig, source_start, source_end = tan_dup_cluster.get_source()
        dest_contig, dest_start, dest_end = tan_dup_cluster.get_destination()
        num_copies = int(round((dest_end - dest_start) / (source_end - source_start)))
        tan_dup_candidates.append(CandidateDuplicationTandem(tan_dup_cluster.source_contig, tan_dup_cluster.source_start, tan_dup_cluster.source_end, num_copies, tan_dup_cluster.members, tan_dup_cluster.score, tan_dup_cluster.std_span, tan_dup_cluster.std_pos))

    ###################################
    # Merge translocation breakpoints #
    ###################################

    # Cluster translocations by contig and pos1
    logging.info("Cluster translocation breakpoints..")
    translocations_fwdfwd = [tra for tra in completed_translocations if tra.direction1 == "fwd" and tra.direction2 == "fwd"]
    translocations_revrev = [tra for tra in completed_translocations if tra.direction1 == "rev" and tra.direction2 == "rev"]
    translocation_partitions_fwdfwd = form_partitions(translocations_fwdfwd, options.trans_partition_max_distance)
    translocation_partitions_revrev = form_partitions(translocations_revrev, options.trans_partition_max_distance)

    translocation_partitions_fwdfwd_dict = defaultdict(list)
    translocation_partitions_revrev_dict = defaultdict(list)
    for partition in translocation_partitions_fwdfwd:
        translocation_partitions_fwdfwd_dict[partition[0].contig1].append(partition)
    for partition in translocation_partitions_revrev:
        translocation_partitions_revrev_dict[partition[0].contig1].append(partition)

    translocation_partition_means_fwdfwd_dict = {}
    translocation_partition_stds_fwdfwd_dict = {}
    for contig in translocation_partitions_fwdfwd_dict.keys():
        translocation_partition_means_fwdfwd_dict[contig] = [int(round(sum([ev.pos1 for ev in partition]) / len(partition))) for partition in translocation_partitions_fwdfwd_dict[contig]]
        translocation_partition_stds_fwdfwd_dict[contig] = [int(round(sqrt(sum([pow(abs(ev.pos1 - translocation_partition_means_fwdfwd_dict[contig][index]), 2) for ev in partition]) / len(partition)))) for index, partition in enumerate(translocation_partitions_fwdfwd_dict[contig])]
    translocation_partition_means_revrev_dict = {}
    translocation_partition_stds_revrev_dict = {}
    for contig in translocation_partitions_revrev_dict.keys():
        translocation_partition_means_revrev_dict[contig] = [int(round(sum([ev.pos1 for ev in partition]) / len(partition))) for partition in translocation_partitions_revrev_dict[contig]]
        translocation_partition_stds_revrev_dict[contig] = [int(round(sqrt(sum([pow(abs(ev.pos1 - translocation_partition_means_revrev_dict[contig][index]), 2) for ev in partition]) / len(partition)))) for index, partition in enumerate(translocation_partitions_revrev_dict[contig])]


    logging.info("Combine inserted regions with translocation breakpoints..")
    new_insertion_from_clusters, inserted_regions_to_remove_1 = merge_translocations_at_insertions(translocation_partitions_fwdfwd_dict, translocation_partition_means_fwdfwd_dict, translocation_partition_stds_fwdfwd_dict, translocation_partitions_revrev_dict, translocation_partition_means_revrev_dict, translocation_partition_stds_revrev_dict, insertion_signature_clusters, options)
    insertion_from_signature_clusters.extend(new_insertion_from_clusters)

    ############################################################################
    # Create interspersed duplication candidates and flag cut&paste insertions #
    ############################################################################

    logging.info("Create interspersed duplication candidates and flag cut&paste insertions..")
    int_duplication_candidates = flag_cutpaste_candidates(insertion_from_signature_clusters, deletion_signature_clusters, options)

    ###################################
    # Remove inserted region clusters #
    ###################################

    #find all inserted regions overlapping interspersed duplication or tandem duplication candidates
    int_duplication_iterator = iter(sorted(int_duplication_candidates, key=lambda cand: cand.get_destination()))
    tan_duplication_iterator = iter(sorted(tan_dup_candidates, key=lambda cand: cand.get_destination()))
    int_duplications_end = False
    tan_duplications_end = False
    inserted_regions_to_remove_2 = []

    try:
        current_int_duplication = next(int_duplication_iterator)
    except StopIteration:
        int_duplications_end = True

    try:
        current_tan_duplication = next(tan_duplication_iterator)
    except StopIteration:
        tan_duplications_end = True

    for inserted_region_index, inserted_region in enumerate(insertion_signature_clusters):
        contig1, start1, end1 = inserted_region.get_source()
        length1 = end1 - start1
        if not int_duplications_end:
            contig2, start2, end2 = current_int_duplication.get_destination()
            while contig2 < contig1 or (contig2 == contig1 and end2 < start1):
                try:
                    current_int_duplication = next(int_duplication_iterator)
                    contig2, start2, end2 = current_int_duplication.get_destination()
                except StopIteration:
                    int_duplications_end = True
                    break
        if not int_duplications_end:
            length2 = end2 - start2
            #if overlapping interspersed duplication of similar length
            if contig2 == contig1 and start2 < end1 and (length1 - length2) / max(length1, length2) < 0.2:
                inserted_regions_to_remove_2.append(inserted_region_index)
        else:
            if not tan_duplications_end:
                contig2, start2, end2 = current_tan_duplication.get_destination()
                while contig2 < contig1 or (contig2 == contig1 and end2 < start1):
                    try:
                        current_tan_duplication = next(tan_duplication_iterator)
                        contig2, start2, end2 = current_tan_duplication.get_destination()
                    except StopIteration:
                        tan_duplications_end = True
                        break
            if not tan_duplications_end:
                length2 = end2 - start2
                #if overlapping tandem duplication of similar length
                if contig2 == contig1 and start2 < end1 and (length1 - length2) / max(length1, length2) < 0.2:
                    inserted_regions_to_remove_2.append(inserted_region_index)

    # remove found inserted regions
    all_inserted_regions_to_remove = sorted(list(set(inserted_regions_to_remove_1 + inserted_regions_to_remove_2)), reverse=True)
    for ins_index in all_inserted_regions_to_remove:
        del(insertion_signature_clusters[ins_index])

    ##############################
    # Create deletion candidates #
    ##############################
    deletion_candidates = []
    for del_cluster in deletion_signature_clusters:
        if del_cluster.score > 0:
            deletion_candidates.append(CandidateDeletion(del_cluster.contig, del_cluster.start, del_cluster.end, del_cluster.members, del_cluster.score, del_cluster.std_span, del_cluster.std_pos))

    #####################################
    # Create novel insertion candidates #
    #####################################
    novel_insertion_candidates = []
    for ins_cluster in insertion_signature_clusters:
        if ins_cluster.score > 0:
            novel_insertion_candidates.append(CandidateNovelInsertion(ins_cluster.contig, ins_cluster.start, ins_cluster.end, ins_cluster.members, ins_cluster.score, ins_cluster.std_span, ins_cluster.std_pos))

    ######################
    # Cluster candidates #
    ######################
    logging.info("Cluster interspersed duplication candidates one more time..")
    final_int_duplication_candidates = cluster_sv_candidates(int_duplication_candidates, options)

    ####################
    # Write candidates #
    ####################
    logging.info("Write SV candidates..")
    logging.info("Final deletion candidates: {0}".format(len(deletion_candidates)))
    logging.info("Final inversion candidates: {0}".format(len(inversion_candidates)))
    logging.info("Final interspersed duplication candidates: {0}".format(len(final_int_duplication_candidates)))
    logging.info("Final tandem duplication candidates: {0}".format(len(tan_dup_candidates)))
    logging.info("Final novel insertion candidates: {0}".format(len(novel_insertion_candidates)))
    write_candidates(working_dir, (final_int_duplication_candidates, inversion_candidates, tan_dup_candidates, deletion_candidates, novel_insertion_candidates))
    write_final_vcf(working_dir, final_int_duplication_candidates, inversion_candidates, tan_dup_candidates, deletion_candidates, novel_insertion_candidates, version, contig_names, contig_lengths, sample)



class Candidate:
    """Candidate class for structural variant candidates. Candidates reflect the final SV types and can be merged from signatures of several reads.
    """
    def __init__(self, source_contig, source_start, source_end, members, score, std_span, std_pos):
        self.source_contig = source_contig
        self.source_start = source_start
        self.source_end = source_end
        
        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.type = "unk"


    def get_source(self):
        return (self.source_contig, self.source_start, self.source_end)


    def get_key(self):
        contig, start, end = self.get_source()
        return (self.type, contig, (start + end) // 2)


    def position_distance_to(self, candidate2):
        """Return position distance between two candidates."""
        this_contig, this_start, this_end = self.get_source()
        other_contig, other_start, other_end = candidate2.get_source()
        this_center = (this_start + this_end) // 2
        other_center = (other_start + other_end) // 2
        if self.type == candidate2.type and this_contig == other_contig:
            return min(abs(this_start - other_start), abs(this_end - other_end), abs(this_center - other_center))
        else:
            return float("inf")


    def get_bed_entry(self):
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}".format(self.source_contig, self.source_start, self.source_end, "{0};{1};{2}".format(self.type, self.std_span, self.std_pos), self.score, ".", "["+"][".join([ev.as_string("|") for ev in self.members])+"]")


    def get_vcf_entry(self):
        raise NotImplementedError


class CandidateDeletion(Candidate):
    def __init__(self, source_contig, source_start, source_end, members, score, std_span, std_pos):
        self.source_contig = source_contig
        #0-based start of the deletion (first deleted base)
        self.source_start = source_start
        #0-based end of the deletion (one past the last deleted base)
        self.source_end = source_end

        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.type = "del"


    def get_vcf_entry(self):
        contig, start, end = self.get_source()
        svtype = "DEL"
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}".format(contig, start, ".", "N", "<" + svtype + ">", int(self.score), "q5" if self.score < 5 else "PASS", "SVTYPE={0};END={1};SVLEN={2};SUPPORT={3};STD_SPAN={4};STD_POS={5}".format(svtype, end, start - end, len(self.members), self.std_span, self.std_pos), "GT", "./.")


class CandidateInversion(Candidate):
    def __init__(self, source_contig, source_start, source_end, members, score, std_span, std_pos):
        self.source_contig = source_contig
        #0-based start of the inversion (first inverted base)
        self.source_start = source_start
        #0-based end of the inversion (one past the last inverted base)
        self.source_end = source_end

        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.type = "inv"


    def get_vcf_entry(self):
        contig, start, end = self.get_source()
        svtype = "INV"
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}".format(contig, start+1, ".", "N", "<" + svtype + ">", int(self.score), "q5" if self.score < 5 else "PASS", "SVTYPE={0};END={1};SUPPORT={2};STD_SPAN={3};STD_POS={4}".format(svtype, end, len(self.members), self.std_span, self.std_pos), "GT", "./.")


class CandidateNovelInsertion(Candidate):
    def __init__(self, dest_contig, dest_start, dest_end, members, score, std_span, std_pos):
        self.dest_contig = dest_contig
        #0-based start of the insertion (base after the insertion)
        self.dest_start = dest_start
        #0-based start of the insertion (base after the insertion) + length of the insertion
        self.dest_end = dest_end

        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.type = "nov_ins"

    def get_destination(self):
        return (self.dest_contig, self.dest_start, self.dest_end)

    def get_bed_entry(self):
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}".format(self.dest_contig, self.dest_start, self.dest_end, "{0};{1};{2}".format(self.type, self.std_span, self.std_pos), self.score, ".", "["+"][".join([ev.as_string("|") for ev in self.members])+"]")

    def get_vcf_entry(self):
        contig, start, end = self.get_destination()
        svtype = "INS:NOVEL"
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}".format(contig, start, ".", "N", "<" + svtype + ">", int(self.score), "q5" if self.score < 5 else "PASS", "SVTYPE={0};END={1};SVLEN={2};SUPPORT={3};STD_SPAN={4};STD_POS={5}".format(svtype, start, end - start, len(self.members), self.std_span, self.std_pos), "GT", "./.")


class CandidateDuplicationTandem(Candidate):
    def __init__(self, source_contig, source_start, source_end, copies, members, score, std_span, std_pos):
        self.source_contig = source_contig
        #0-based start of the region (first copied base)
        self.source_start = source_start
        #0-based end of the region (one past the last copied base)
        self.source_end = source_end
        
        self.copies = copies

        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.type = "dup_tan"


    def get_destination(self):
        source_contig, source_start, source_end = self.get_source()
        return (source_contig, source_end, source_end + self.copies * (source_end - source_start))


    def get_bed_entries(self, sep="\t"):
        source_contig, source_start, source_end = self.get_source()
        dest_contig, dest_start, dest_end = self.get_destination()
        source_entry = sep.join(["{0}", "{1}", "{2}", "{3}", "{4}", "{5}", "{6}"]).format(source_contig, source_start,
                                                                                     source_end,
                                                                                     "tan_dup_source;>{0}:{1}-{2};{3};{4}".format(
                                                                                         dest_contig, dest_start,
                                                                                         dest_end, self.std_span, self.std_pos), self.score, ".",
                                                                                     "[" + "][".join(
                                                                                         [ev.as_string("|") for ev in
                                                                                          self.members]) + "]")
        dest_entry = sep.join(["{0}", "{1}", "{2}", "{3}", "{4}", "{5}", "{6}"]).format(dest_contig, dest_start, dest_end,
                                                                                   "tan_dup_dest;<{0}:{1}-{2};{3};{4}".format(
                                                                                       source_contig, source_start,
                                                                                       source_end, self.std_span, self.std_pos), self.score, ".",
                                                                                   "[" + "][".join(
                                                                                       [ev.as_string("|") for ev in
                                                                                        self.members]) + "]")
        return (source_entry, dest_entry)


    def get_vcf_entry(self):
        contig = self.source_contig
        start = self.source_end
        end = self.source_end + self.copies * (self.source_end - self.source_start)
        svtype = "DUP:TANDEM"
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}".format(contig, start, ".", "N", "<" + svtype + ">", int(self.score), "q5" if self.score < 5 else "PASS", "SVTYPE={0};END={1};SVLEN={2};SUPPORT={3};STD_SPAN={4};STD_POS={5}".format(svtype, start, end - start, len(self.members), self.std_span, self.std_pos), "GT", "./.")


class CandidateDuplicationInterspersed(Candidate):
    def __init__(self, source_contig, source_start, source_end, dest_contig, dest_start, dest_end, members, score, std_span, std_pos, cutpaste=False):
        self.source_contig = source_contig
        #0-based start of the region (first copied base)
        self.source_start = source_start
        #0-based end of the region (one past the last copied base)
        self.source_end = source_end

        self.dest_contig = dest_contig
        #0-based start of the insertion (base after the insertion)
        self.dest_start = dest_start
        #0-based end of the insertion (base after the insertion) + length of the insertion
        self.dest_end = dest_end

        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.cutpaste= cutpaste
        self.type = "dup_int"


    def get_destination(self):
        return (self.dest_contig, self.dest_start, self.dest_end)


    def get_bed_entries(self, sep="\t"):
        source_contig, source_start, source_end = self.get_source()
        dest_contig, dest_start, dest_end = self.get_destination()
        source_entry = sep.join(["{0}", "{1}", "{2}", "{3}", "{4}", "{5}", "{6}"]).format(source_contig, source_start,
                                                                                     source_end,
                                                                                     "int_dup_source;>{0}:{1}-{2};{3};{4}".format(
                                                                                         dest_contig, dest_start,
                                                                                         dest_end, self.std_span, self.std_pos), self.score, "origin potentially deleted" if self.cutpaste else ".",
                                                                                     "[" + "][".join(
                                                                                         [ev.as_string("|") for ev in
                                                                                          self.members]) + "]")
        dest_entry = sep.join(["{0}", "{1}", "{2}", "{3}", "{4}", "{5}", "{6}"]).format(dest_contig, dest_start, dest_end,
                                                                                   "int_dup_dest;<{0}:{1}-{2};{3};{4}".format(
                                                                                       source_contig, source_start,
                                                                                       source_end, self.std_span, self.std_pos), self.score, "origin potentially deleted" if self.cutpaste else ".",
                                                                                   "[" + "][".join(
                                                                                       [ev.as_string("|") for ev in
                                                                                        self.members]) + "]")
        return (source_entry, dest_entry)


    def get_vcf_entry(self):
        contig, start, end = self.get_destination()
        svtype = "DUP:INT"
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}".format(contig, start, ".", "N", "<" + svtype + ">", int(self.score), "q5" if self.score < 5 else "PASS", "SVTYPE={0};{1}END={2};SVLEN={3};SUPPORT={4};STD_SPAN={5};STD_POS={6}".format(svtype, "CUTPASTE;" if self.cutpaste else "", start, end - start, len(self.members), self.std_span, self.std_pos), "GT", "./.")



import unittest
import tempfile

from svim.SVIM_input_parsing import guess_file_type, parse_arguments

class TestInputParsing(unittest.TestCase):

    def test_guess_file_type(self):
        fasta_paths = ["/test/path/test.file.fa", "/test/path/test.file.fasta", "/test/path/test.file.FA"]
        fastq_paths = ["/test/path/test.file.fq", "/test/path/test.file.FQ", "/test/path/test.file.fastq"]

        for p in fasta_paths:
            self.assertEqual(guess_file_type(p), "fasta")
            zipped_path1 = p + ".gz"
            zipped_path2 = p + ".gzip"
            self.assertEqual(guess_file_type(zipped_path1), "fasta_gzip")
            self.assertEqual(guess_file_type(zipped_path2), "fasta_gzip")
            list_path = p + ".fn"
            self.assertEqual(guess_file_type(list_path), "list")
        
        for p in fastq_paths:
            self.assertEqual(guess_file_type(p), "fastq")
            zipped_path1 = p + ".gz"
            zipped_path2 = p + ".gzip"
            self.assertEqual(guess_file_type(zipped_path1), "fastq_gzip")
            self.assertEqual(guess_file_type(zipped_path2), "fastq_gzip")
            list_path = p + ".fn"
            self.assertEqual(guess_file_type(list_path), "list")
    
    def test_parse_arguments(self):
        arguments = ['alignment', 'myworkdir', 'mybamfile']
        options = parse_arguments('0.4.3', arguments)
        self.assertEqual(type(options.min_mapq), int)
        self.assertEqual(type(options.min_sv_size), int)
        self.assertEqual(type(options.max_sv_size), int)
        self.assertEqual(type(options.skip_indel), bool)
        self.assertEqual(type(options.skip_segment), bool)
        self.assertEqual(type(options.segment_gap_tolerance), int)
        self.assertEqual(type(options.segment_overlap_tolerance), int)
        self.assertEqual(type(options.partition_max_distance), int)
        self.assertEqual(type(options.distance_normalizer), int)
        self.assertEqual(type(options.cluster_max_distance), float)
        self.assertEqual(type(options.del_ins_dup_max_distance), float)
        self.assertEqual(type(options.trans_destination_partition_max_distance), int)
        self.assertEqual(type(options.trans_partition_max_distance), int)
        self.assertEqual(type(options.trans_sv_max_distance), int)
        self.assertEqual(type(options.sample), str)
        self.assertEqual(type(options.trans_partition_max_distance), int)
        self.assertEqual(type(options.trans_partition_max_distance), int)



import logging
import pysam

from svim.SVIM_intra import analyze_alignment_indel
from svim.SVIM_inter import analyze_read_segments


def bam_iterator(bam):
    """Returns an iterator for the given SAM/BAM file (must be query-sorted).
    In each call, the alignments of a single read are yielded as a 3-tuple: (list of primary pysam.AlignedSegment, list of supplementary pysam.AlignedSegment, list of secondary pysam.AlignedSegment)."""
    alignments = bam.fetch(until_eof=True)
    current_aln = next(alignments)
    current_read_name = current_aln.query_name
    current_prim = []
    current_suppl = []
    current_sec = []
    if current_aln.is_secondary:
        current_sec.append(current_aln)
    elif current_aln.is_supplementary:
        current_suppl.append(current_aln)
    else:
        current_prim.append(current_aln)
    while True:
        try:
            next_aln = next(alignments)
            next_read_name = next_aln.query_name
            if next_read_name != current_read_name:
                yield (current_prim, current_suppl, current_sec)
                current_read_name = next_read_name
                current_prim = []
                current_suppl = []
                current_sec = []
            if next_aln.is_secondary:
                current_sec.append(next_aln)
            elif next_aln.is_supplementary:
                current_suppl.append(next_aln)
            else:
                current_prim.append(next_aln)
        except StopIteration:
            break
    yield (current_prim, current_suppl, current_sec)


def retrieve_supplementary_alignments(primary_alignment, bam):
    """Reconstruct supplementary alignments for a given primary alignment from the SA tag"""
    try:
        sa_tag = primary_alignment.get_tag("SA").split(";")           
    except KeyError:
        return []
    supplementary_alignments = []
    # For each supplementary alignment encoded in the SA tag
    for element in sa_tag:
        # Read information from the tag
        fields = element.split(",")
        if len(fields) != 6:
            continue
        rname = fields[0]
        pos = int(fields[1])
        strand = fields[2]
        # CIGAR string encoded in SA tag is shortened
        cigar = fields[3]
        mapq = int(fields[4])
        nm = int(fields[5])

        # Generate an aligned segment from the information
        a = pysam.AlignedSegment()
        a.query_name = primary_alignment.query_name
        a.query_sequence= primary_alignment.query_sequence
        if strand == "+":
            a.flag = 2048
        else:
            a.flag = 2064
        a.reference_id = bam.get_tid(rname)
        a.reference_start = pos - 1
        try:
            a.mapping_quality = mapq
        except OverflowError:
            a.mapping_quality = 0
        a.cigarstring = cigar
        a.next_reference_id = -1
        a.next_reference_start = -1
        a.template_length = 0
        a.query_qualities = primary_alignment.query_qualities
        a.set_tags([("NM", nm, "i")])

        supplementary_alignments.append(a)
    return supplementary_alignments


def analyze_alignment_file_querysorted(bam, options):
    alignment_it = bam_iterator(bam)

    sv_signatures = []
    read_nr = 0

    while True:
        try:
            alignment_iterator_object = next(alignment_it)
            primary_aln, suppl_aln, sec_aln = alignment_iterator_object
            if len(primary_aln) != 1 or primary_aln[0].is_unmapped or primary_aln[0].mapping_quality < options.min_mapq:
                continue
            read_nr += 1
            if read_nr % 10000 == 0:
                logging.info("Processed read {0}".format(read_nr))
            good_suppl_alns = [aln for aln in suppl_aln if not aln.is_unmapped and aln.mapping_quality >= options.min_mapq]
            if not options.skip_indel:
                sv_signatures.extend(analyze_alignment_indel(primary_aln[0], bam, primary_aln[0].query_name, options))
                for alignment in good_suppl_alns:
                    sv_signatures.extend(analyze_alignment_indel(alignment, bam, alignment.query_name, options))
            if not options.skip_segment:
                sv_signatures.extend(analyze_read_segments(primary_aln[0], good_suppl_alns, bam, options))
        except StopIteration:
            break
        except KeyboardInterrupt:
            logging.warning('Execution interrupted by user. Stop detection and continue with next step..')
            break
    return sv_signatures


def analyze_alignment_file_coordsorted(bam, options):
    alignment_it = bam.fetch(until_eof=True)

    sv_signatures = []
    read_nr = 0

    while True:
        try:
            current_alignment = next(alignment_it)
            if current_alignment.is_unmapped or current_alignment.is_secondary or current_alignment.mapping_quality < options.min_mapq:
                continue
            if current_alignment.is_supplementary:
                if not options.skip_indel:
                    sv_signatures.extend(analyze_alignment_indel(current_alignment, bam, current_alignment.query_name, options))
            else:
                read_nr += 1
                if read_nr % 10000 == 0:
                    logging.info("Processed read {0}".format(read_nr))
                supplementary_alignments = retrieve_supplementary_alignments(current_alignment, bam)
                good_suppl_alns = [aln for aln in supplementary_alignments if not aln.is_unmapped and aln.mapping_quality >= options.min_mapq]

                if not options.skip_indel:
                    sv_signatures.extend(analyze_alignment_indel(current_alignment, bam, current_alignment.query_name, options))
                if not options.skip_segment:
                    sv_signatures.extend(analyze_read_segments(current_alignment, good_suppl_alns, bam, options))
        except StopIteration:
            break
        except KeyboardInterrupt:
            logging.warning('Execution interrupted by user. Stop detection and continue with next step..')
            break
    return sv_signatures


import unittest

from svim.SVSignature import SignatureDeletion,SignatureInsertion

class TestSVSignature(unittest.TestCase):

    def test_accessors(self):
        deletion = SignatureDeletion("chr1", 100, 300, "cigar", "read1")

        self.assertEqual(deletion.get_source(), ("chr1", 100, 300))
        self.assertEqual(deletion.get_key(), ("del", "chr1", 200))
    
    def test_position_distance_to(self):
        deletion1 = SignatureDeletion("chr1", 100, 300, "cigar", "read1")
        deletion2 = SignatureDeletion("chr1", 150, 200, "cigar", "read2")
        deletion3 = SignatureDeletion("chr2", 150, 200, "cigar", "read2")
        insertion = SignatureInsertion("chr1", 150, 200, "cigar", "read2")

        self.assertEqual(deletion1.position_distance_to(deletion2), 25)
        self.assertEqual(deletion1.position_distance_to(deletion3), float("Inf"))
        self.assertEqual(deletion1.position_distance_to(insertion), float("Inf"))
    
    def test_as_string(self):
        deletion1 = SignatureDeletion("chr1", 100, 300, "cigar", "read1")

        self.assertEqual(deletion1.as_string(), "chr1\t100\t300\tdel;cigar\tread1")
        self.assertEqual(deletion1.as_string(":"), "chr1:100:300:del;cigar:read1")


if __name__ == '__main__':
    unittest.main()



import os
import logging

from subprocess import run, CalledProcessError

class ToolMissingError(Exception): pass

class AlignmentPipelineError(Exception): pass


def check_prereqisites(aligner):
    devnull = open(os.devnull, 'w')
    try:
        run(['gunzip', '--help'], stdout=devnull, stderr=devnull, check=True)
        run([aligner, '--help'], stdout=devnull, stderr=devnull, check=True)
        run(['samtools', '--help'], stdout=devnull, stderr=devnull, check=True)
    except FileNotFoundError as e:
        raise ToolMissingError('The alignment pipeline cannot be started because {0} was not found. Is it installed and in the PATH?'.format(e.filename)) from e
    except CalledProcessError as e:
        raise ToolMissingError('The alignment pipeline cannot be started because {0} failed.'.format(" ".join(e.cmd))) from e


def run_alignment(working_dir, genome, reads_path, reads_type, cores, aligner, nanopore):
    """Align full reads with NGMLR or minimap2."""
    check_prereqisites(aligner)
    reads_file_prefix = os.path.splitext(os.path.basename(reads_path))[0]
    full_aln = "{0}/{1}.{2}.querysorted.bam".format(working_dir, reads_file_prefix, aligner)
    if not os.path.exists(full_aln):
        try:
            command = ['set', '-o', 'pipefail', '&&']
            if aligner == "ngmlr":
                # We need to uncompress gzipped files for NGMLR first
                if reads_type == "fasta_gzip" or reads_type == "fastq_gzip":
                    command += ['gunzip', '-c', os.path.realpath(reads_path)]
                    command += ['|', 'ngmlr', '-t', str(cores), '-r', genome]
                    if nanopore:
                        command += ['-x', 'ont']
                else:
                    command += ['ngmlr', '-t', str(cores), '-r', genome, '-q', os.path.realpath(reads_path)]
                    if nanopore:
                        command += ['-x', 'ont']
            elif aligner == "minimap2":
                if nanopore:
                    command += ['minimap2', '-t', str(cores), '-x', 'map-ont', '-a', genome, os.path.realpath(reads_path)]
                else:
                    command += ['minimap2', '-t', str(cores), '-x', 'map-pb', '-a', genome, os.path.realpath(reads_path)]
            command += ['|', 'samtools', 'view', '-b', '-@', str(cores)]
            command += ['|', 'samtools', 'sort', '-n', '-@', str(cores), '-o', full_aln]
            logging.info("Starting alignment pipeline..")
            run(" ".join(command), shell=True, check=True)
        except CalledProcessError as e:
            raise AlignmentPipelineError('The alignment pipeline failed with exit code {0}. Command was: {1}'.format(e.returncode, e.cmd)) from e
        logging.info("Alignment pipeline finished")
        return full_aln
    else:
        logging.warning("Alignment output file {0} already exists. Skip alignment and use the existing file.".format(full_aln))
        return full_aln


import unittest
import pysam
import os

from svim.SVIM_COLLECT import retrieve_supplementary_alignments

class TestSAExtraction(unittest.TestCase):
    def setUp(self):
        TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), 'chimeric_read.bam')
        self.samfile = pysam.AlignmentFile(TESTDATA_FILENAME, "rb")
        self.alignments = list(self.samfile.fetch(until_eof=True))

    def test_fetch(self):
        self.assertEqual(len(self.alignments), 4)
    
    def test_satag_length(self):
        primary = self.alignments[0]
        supplementary_alns = retrieve_supplementary_alignments(primary, self.samfile)
        self.assertEqual(len(supplementary_alns), 3)
    
    def test_satag_extraction_complete(self):
        primary = self.alignments[0]
        supplementary_alns = retrieve_supplementary_alignments(primary, self.samfile)
        for index, aln in enumerate(supplementary_alns):
            self.assertEqual(aln.cigarstring, self.alignments[index+1].cigarstring)
            self.assertEqual(aln.reference_id, self.alignments[index+1].reference_id)
            self.assertEqual(aln.reference_start, self.alignments[index+1].reference_start)
            self.assertEqual(aln.reference_end, self.alignments[index+1].reference_end)
            self.assertEqual(aln.flag, self.alignments[index+1].flag)
            self.assertEqual(aln.mapping_quality, self.alignments[index+1].mapping_quality)
            self.assertEqual(aln.query_sequence, self.alignments[index+1].query_sequence)
            self.assertEqual(aln.query_name, self.alignments[index+1].query_name)
            self.assertEqual(aln.query_alignment_start, self.alignments[index+1].query_alignment_start)
            self.assertEqual(aln.query_alignment_end, self.alignments[index+1].query_alignment_end)


class TestSAExtractionError(unittest.TestCase):
    def setUp(self):
        TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), 'chimeric_read_errors.bam')
        self.samfile = pysam.AlignmentFile(TESTDATA_FILENAME, "rb")
        self.alignments = list(self.samfile.fetch(until_eof=True))
    
    def test_satag_too_many_fields(self):
        primary = self.alignments[0]
        supplementary_alns = retrieve_supplementary_alignments(primary, self.samfile)
        #first SA entry has too many fields
        self.assertEqual(len(supplementary_alns), 2)

    def test_satag_negative_mapq(self):
        primary = self.alignments[1]
        supplementary_alns = retrieve_supplementary_alignments(primary, self.samfile)
        #negative mapping quality is interpreted as 0
        self.assertEqual(len(supplementary_alns), 1)
        self.assertEqual(supplementary_alns[0].mapping_quality, 0)

if __name__ == '__main__':
    unittest.main()


from __future__ import print_function

import sys

from svim.SVSignature import SignatureDeletion, SignatureInsertion


def analyze_cigar_indel(tuples, min_length):
    """Parses CIGAR tuples (op, len) and returns Indels with a length > minLength"""
    pos = 0
    indels = []
    for operation, length in tuples:
        if operation == 0:                     # alignment match
            pos += length
        elif operation == 1:                   # insertion
            if length >= min_length:
                indels.append((pos, length, 'ins'))
        elif operation == 2:                   # deletion
            if length >= min_length:
                indels.append((pos, length, 'del'))
            pos += length
        elif operation == 7 or operation == 8:        # match or mismatch
            pos += length
    return indels


def analyze_alignment_indel(alignment, bam, query_name, options):
    sv_signatures = []
    ref_chr = bam.getrname(alignment.reference_id)
    ref_start = alignment.reference_start
    indels = analyze_cigar_indel(alignment.cigartuples, options.min_sv_size)
    for pos, length, typ in indels:
        if typ == "del":
            sv_signatures.append(SignatureDeletion(ref_chr, ref_start + pos, ref_start + pos + length, "cigar", query_name))
        elif typ == "ins":
            sv_signatures.append(SignatureInsertion(ref_chr, ref_start + pos, ref_start + pos + length, "cigar", query_name))
    return sv_signatures





import unittest
import pysam
import tempfile

from svim.SVIM_COLLECT import bam_iterator, analyze_alignment_file_querysorted, retrieve_supplementary_alignments
from svim.SVIM_input_parsing import parse_arguments
from random import choice, triangular, uniform

class TestCollect(unittest.TestCase):

    def generate_random_sequence(self, length):
        sequence=""
        for i in range(length):
            sequence+=choice("ACGT")
        return sequence

    def generate_random_cigar_string(self, readlength):
        """Generate random cigar string for a read of a given length. Simulate small mismatches and indels but nothing larger than 10bp."""
        softclip_left = round(triangular(0, readlength, min(1000, readlength * 0.5)))
        non_clipped = readlength - softclip_left
        softclip_right = round(triangular(0, non_clipped, min(1000, non_clipped * 0.5)))
        non_clipped = readlength - softclip_left - softclip_right
        sequence = ""
        read_bases_consumed = 0
        while read_bases_consumed < non_clipped:
            #choose next operation
            if len(sequence) == 0 or sequence[-1] == "I" or sequence[-1] == "D":
                next_operation = "M"
                next_length = round(triangular(1, non_clipped - read_bases_consumed, min(30, non_clipped - read_bases_consumed)))
                read_bases_consumed += next_length
            else:
                next_operation = choice("ID")
                if next_operation == "I":
                    next_length = round(triangular(1, min(10, non_clipped - read_bases_consumed), 1))
                    read_bases_consumed += next_length
                else:
                    next_length = round(triangular(1, 10, 1))
            sequence += str(next_length) + next_operation
        return "{0}S{1}{2}S".format(softclip_left, sequence, softclip_right)

    def generate_read(self, qname, flag):
        rname = "chr1"
        pos = int(uniform(1,249250620))
        mapq = int(triangular(0, 60, 50))
        length = int(triangular(100, 20000, 15000))
        cigar = self.generate_random_cigar_string(length)
        seq = self.generate_random_sequence(length)

        read_info = (qname, flag, rname, pos, mapq, cigar, "*", 0, 0, seq, "*", "")

        return read_info
    
    def generate_primary_read_with_sa_tag(self, qname, flag):
        rname = "chr1"
        pos = int(uniform(1,249250620))
        mapq = int(triangular(0, 60, 50))
        length = int(triangular(100, 20000, 15000))
        cigar = self.generate_random_cigar_string(length)
        seq = self.generate_random_sequence(length)

        supplementary_read_info = (qname, flag + 2048, rname, pos, mapq, cigar, "*", 0, 0, seq, "*", "")
        sa_tag = "SA:Z:{rname},{pos},{strand},{cigar},{mapq},{nm};".format(rname=rname, pos=pos, strand=("-" if flag & 16 else "+"), cigar=cigar, mapq=mapq, nm=0)
        
        rname = "chr1"
        pos = int(uniform(1,249250620))
        mapq = int(triangular(0, 60, 50))
        cigar = self.generate_random_cigar_string(length)

        primary_read_info = (qname, flag, rname, pos, mapq, cigar, "*", 0, 0, seq, "*", sa_tag)

        return (primary_read_info, supplementary_read_info)

    def setUp(self):
        self.bam_file = tempfile.NamedTemporaryFile()
        header = """@HD	VN:1.0	SO:queryname
@SQ	SN:chr1	LN:249250621
@SQ	SN:chr2	LN:243199373
@SQ	SN:chr3	LN:198022430
@SQ	SN:chr4	LN:191154276
@SQ	SN:chr5	LN:180915260
@SQ	SN:chr6	LN:171115067
@SQ	SN:chr7	LN:159138663
@SQ	SN:chr8	LN:146364022
@SQ	SN:chr9	LN:141213431
@SQ	SN:chr10	LN:135534747
@SQ	SN:chr11	LN:135006516
@SQ	SN:chr12	LN:133851895
@SQ	SN:chr13	LN:115169878
@SQ	SN:chr14	LN:107349540
@SQ	SN:chr15	LN:102531392
@SQ	SN:chr16	LN:90354753
@SQ	SN:chr17	LN:81195210
@SQ	SN:chr18	LN:78077248
@SQ	SN:chr19	LN:59128983
@SQ	SN:chr20	LN:63025520
@SQ	SN:chr21	LN:48129895
@SQ	SN:chr22	LN:51304566
@SQ	SN:chrX	LN:155270560
@SQ	SN:chrY	LN:59373566
@SQ	SN:chr6_ssto_hap7	LN:4928567
@SQ	SN:chr6_mcf_hap5	LN:4833398
@SQ	SN:chr6_cox_hap2	LN:4795371
@SQ	SN:chr6_mann_hap4	LN:4683263
@SQ	SN:chr6_apd_hap1	LN:4622290
@SQ	SN:chr6_qbl_hap6	LN:4611984
@SQ	SN:chr6_dbb_hap3	LN:4610396
@SQ	SN:chr17_ctg5_hap1	LN:1680828
@SQ	SN:chr4_ctg9_hap1	LN:590426
@SQ	SN:chr1_gl000192_random	LN:547496
@SQ	SN:chrUn_gl000225	LN:211173
@SQ	SN:chr4_gl000194_random	LN:191469
@SQ	SN:chr4_gl000193_random	LN:189789
@SQ	SN:chr9_gl000200_random	LN:187035
@SQ	SN:chrUn_gl000222	LN:186861
@SQ	SN:chrUn_gl000212	LN:186858
@SQ	SN:chr7_gl000195_random	LN:182896
@SQ	SN:chrUn_gl000223	LN:180455
@SQ	SN:chrUn_gl000224	LN:179693
@SQ	SN:chrUn_gl000219	LN:179198
@SQ	SN:chr17_gl000205_random	LN:174588
@SQ	SN:chrUn_gl000215	LN:172545
@SQ	SN:chrUn_gl000216	LN:172294
@SQ	SN:chrUn_gl000217	LN:172149
@SQ	SN:chr9_gl000199_random	LN:169874
@SQ	SN:chrUn_gl000211	LN:166566
@SQ	SN:chrUn_gl000213	LN:164239
@SQ	SN:chrUn_gl000220	LN:161802
@SQ	SN:chrUn_gl000218	LN:161147
@SQ	SN:chr19_gl000209_random	LN:159169
@SQ	SN:chrUn_gl000221	LN:155397
@SQ	SN:chrUn_gl000214	LN:137718
@SQ	SN:chrUn_gl000228	LN:129120
@SQ	SN:chrUn_gl000227	LN:128374
@SQ	SN:chr1_gl000191_random	LN:106433
@SQ	SN:chr19_gl000208_random	LN:92689
@SQ	SN:chr9_gl000198_random	LN:90085
@SQ	SN:chr17_gl000204_random	LN:81310
@SQ	SN:chrUn_gl000233	LN:45941
@SQ	SN:chrUn_gl000237	LN:45867
@SQ	SN:chrUn_gl000230	LN:43691
@SQ	SN:chrUn_gl000242	LN:43523
@SQ	SN:chrUn_gl000243	LN:43341
@SQ	SN:chrUn_gl000241	LN:42152
@SQ	SN:chrUn_gl000236	LN:41934
@SQ	SN:chrUn_gl000240	LN:41933
@SQ	SN:chr17_gl000206_random	LN:41001
@SQ	SN:chrUn_gl000232	LN:40652
@SQ	SN:chrUn_gl000234	LN:40531
@SQ	SN:chr11_gl000202_random	LN:40103
@SQ	SN:chrUn_gl000238	LN:39939
@SQ	SN:chrUn_gl000244	LN:39929
@SQ	SN:chrUn_gl000248	LN:39786
@SQ	SN:chr8_gl000196_random	LN:38914
@SQ	SN:chrUn_gl000249	LN:38502
@SQ	SN:chrUn_gl000246	LN:38154
@SQ	SN:chr17_gl000203_random	LN:37498
@SQ	SN:chr8_gl000197_random	LN:37175
@SQ	SN:chrUn_gl000245	LN:36651
@SQ	SN:chrUn_gl000247	LN:36422
@SQ	SN:chr9_gl000201_random	LN:36148
@SQ	SN:chrUn_gl000235	LN:34474
@SQ	SN:chrUn_gl000239	LN:33824
@SQ	SN:chr21_gl000210_random	LN:27682
@SQ	SN:chrUn_gl000231	LN:27386
@SQ	SN:chrUn_gl000229	LN:19913
@SQ	SN:chrM	LN:16571
@SQ	SN:chrUn_gl000226	LN:15008
@SQ	SN:chr18_gl000207_random	LN:4262
@PG	ID:ngmlr	PN:nextgenmap-lr	VN:0.2.7	CL:ngmlr -t 10 -r hg19.fa -q reads.fa -o reads.ngmlr.hg19.bam"""
        self.bam_file.write(header.encode('utf-8'))

        self.read_infos = []
        #10 reads with only primary alignment
        for index in range(10):
            read_info = self.generate_read("read{}".format(index+1), 0)
            self.read_infos.append(read_info)
            sam_entry = "\n" + "\t".join([str(el) for el in read_info])
            self.bam_file.write(sam_entry.encode('utf-8'))

        #10 reads with primary and supplementary alignment
        for index in range(10, 20):
            primary_read_info, supplementary_read_info = self.generate_primary_read_with_sa_tag("read{}".format(index+1), 0)
            #primary with SA tag
            self.read_infos.append(primary_read_info)
            sam_entry = "\n" + "\t".join([str(el) for el in primary_read_info])
            self.bam_file.write(sam_entry.encode('utf-8'))
            #supplementary
            self.read_infos.append(supplementary_read_info)
            sam_entry = "\n" + "\t".join([str(el) for el in supplementary_read_info])
            self.bam_file.write(sam_entry.encode('utf-8'))            

        self.bam_file.seek(0)
        self.alignment_file = pysam.AlignmentFile(self.bam_file.name, "rb")

    def test_bam_iterator(self):
        bam_it = bam_iterator(self.alignment_file)

        num_primary_only = 0
        num_primary_supplementary = 0
        num_total = 0
        for prim, suppl, sec in bam_it:
            if len(prim) == 1 and len(suppl) == 0 and len(sec) == 0:
                num_primary_only += 1                
            if len(prim) == 1 and len(suppl) == 1 and len(sec) == 0:
                num_primary_supplementary += 1
            num_total += 1
        
        self.assertEqual(num_total, 20)
        self.assertEqual(num_primary_only, 10)
        self.assertEqual(num_primary_supplementary, 10)

    def test_analyze_alignment_file_querysorted(self):
        arguments = ['alignment', '--skip_segment', 'myworkdir', 'mybamfile']
        options = parse_arguments('0.4.3', arguments)
        signatures = analyze_alignment_file_querysorted(self.alignment_file, options)

        self.assertEqual(len(signatures), 0)
    
    def test_retrieve_supplementary_alignments(self):
        alignment_it = self.alignment_file.fetch(until_eof=True)
        alignments = list(alignment_it)
        for i in range(10,30,2):
            primary = alignments[i]
            supplementary = alignments[i+1]
            retrieved_supplementary_alns = retrieve_supplementary_alignments(primary, self.alignment_file)
            self.assertEqual(len(retrieved_supplementary_alns), 1)
            self.assertEqual(retrieved_supplementary_alns[0].cigarstring, supplementary.cigarstring)
            self.assertEqual(retrieved_supplementary_alns[0].reference_id, supplementary.reference_id)
            self.assertEqual(retrieved_supplementary_alns[0].reference_start, supplementary.reference_start)
            self.assertEqual(retrieved_supplementary_alns[0].reference_end, supplementary.reference_end)
            self.assertEqual(retrieved_supplementary_alns[0].flag, supplementary.flag)
            self.assertEqual(retrieved_supplementary_alns[0].mapping_quality, supplementary.mapping_quality)
            self.assertEqual(retrieved_supplementary_alns[0].query_sequence, supplementary.query_sequence)
            self.assertEqual(retrieved_supplementary_alns[0].query_name, supplementary.query_name)
            self.assertEqual(retrieved_supplementary_alns[0].query_alignment_start, supplementary.query_alignment_start)
            self.assertEqual(retrieved_supplementary_alns[0].query_alignment_end, supplementary.query_alignment_end)




from __future__ import print_function

import sys
import logging

from random import sample
from statistics import mean, stdev
import numpy as np
from scipy.cluster.hierarchy import linkage, fcluster

from svim.SVSignature import SignatureClusterUniLocal, SignatureClusterBiLocal
from svim.SVCandidate import CandidateDuplicationInterspersed


def form_partitions(sv_signatures, max_distance):
    """Form partitions of signatures using mean distance."""
    sorted_signatures = sorted(sv_signatures, key=lambda evi: evi.get_key())
    partitions = []
    current_partition = []
    for signature in sorted_signatures:
        if len(current_partition) > 0 and current_partition[-1].position_distance_to(signature) > max_distance:
            partitions.append(current_partition[:])
            current_partition = []
        current_partition.append(signature)
    if len(current_partition) > 0:
        partitions.append(current_partition[:])
    return partitions


def span_position_distance(signature1, signature2):
    distance_normalizer = signature1[2]
    span1 = signature1[1] - signature1[0]
    span2 = signature2[1] - signature2[0]
    center1 = (signature1[0] + signature1[1]) // 2
    center2 = (signature2[0] + signature2[1]) // 2
    position_distance = min(abs(signature1[0] - signature2[0]), abs(signature1[1] - signature2[1]), abs(center1 - center2)) / distance_normalizer
    span_distance = abs(span1 - span2) / max(span1, span2)
    return position_distance + span_distance


def span_position_distance_bilocal(signature1, signature2):
    "Special span position distance function for intersepersed duplication signatures"
    distance_normalizer = signature1[3]
    span1 = signature1[1] - signature1[0]
    span2 = signature2[1] - signature2[0]
    source_center1 = (signature1[0] + signature1[1]) // 2
    source_center2 = (signature2[0] + signature2[1]) // 2
    position_distanc_source = min(abs(signature1[0] - signature2[0]), abs(signature1[1] - signature2[1]), abs(source_center1 - source_center2)) / distance_normalizer
    position_distanc_destination = abs(signature1[2] - signature2[2]) / distance_normalizer
    span_distance = abs(span1 - span2) / max(span1, span2)
    return position_distanc_source + position_distanc_destination + span_distance


def clusters_from_partitions(partitions, options):
    """Finds clusters in partitions using span-position distance and hierarchical clustering. 
    Assumes that all signatures in the given partition are of the same type and on the same contig"""
    clusters_final = []
    # Find clusters in each partition individually.
    for partition in partitions:
        if len(partition) == 1:
            clusters_final.append([partition[0]])
            continue
        elif len(partition) > 100:
            partition_sample = sample(partition, 100)
        else:
            partition_sample = partition
        element_type = partition_sample[0].type
        #Uni-local clustering
        if element_type == "del" or element_type == "ins" or element_type == "inv" or element_type == "dup":
            data = np.array( [[signature.get_source()[1], signature.get_source()[2], options.distance_normalizer] for signature in partition_sample])
            Z = linkage(data, method = "average", metric = span_position_distance)
        #Bi-local clustering
        elif element_type == "ins_dup" or element_type == "dup_int":
            data = np.array( [[signature.get_source()[1], signature.get_source()[2], signature.get_destination()[1], options.distance_normalizer] for signature in partition_sample])
            Z = linkage(data, method = "average", metric = span_position_distance_bilocal)

        cluster_indices = list(fcluster(Z, options.cluster_max_distance, criterion='distance'))
        new_clusters = [[] for i in range(max(cluster_indices))]
        for signature_index, cluster_index in enumerate(cluster_indices):
            new_clusters[cluster_index-1].append(partition_sample[signature_index])
        clusters_final.extend(new_clusters)
    return clusters_final


def calculate_score(cluster, std_span, std_pos, span, type):
    if std_span == None or std_pos == None:
        span_deviation_score = 0
        pos_deviation_score = 0
    else:
        span_deviation_score = 1 - min(1, std_span / span)
        pos_deviation_score = 1 - min(1, std_pos / span)

    if type == "inv":
        directions = [signature.direction for signature in cluster]
        direction_counts = [0, 0, 0, 0, 0]
        for direction in directions:
            if direction == "left_fwd":
                direction_counts[0] += 1
            if direction == "left_rev":
                direction_counts[1] += 1
            if direction == "right_fwd":
                direction_counts[2] += 1
            if direction == "right_rev":
                direction_counts[3] += 1
            if direction == "all":
                direction_counts[4] += 1
        left_signatures = direction_counts[0] + direction_counts[1]
        right_signatures = direction_counts[2] + direction_counts[3]
        valid_signatures = min(left_signatures, right_signatures) + direction_counts[4]
        num_signatures = min(80, valid_signatures)
    else:
        num_signatures = min(80, len(cluster))
    return num_signatures + span_deviation_score * (num_signatures / 8) + pos_deviation_score * (num_signatures / 8)


def consolidate_clusters_unilocal(clusters):
    """Consolidate clusters to a list of (type, contig, mean start, mean end, cluster size, members) tuples."""
    consolidated_clusters = []
    for cluster in clusters:
        average_start = sum([member.get_source()[1] for member in cluster]) / len(cluster)
        average_end = sum([member.get_source()[2] for member in cluster]) / len(cluster)
        if len(cluster) > 1:
            std_span = stdev([member.get_source()[2] - member.get_source()[1] for member in cluster])
            std_pos = stdev([(member.get_source()[2] + member.get_source()[1]) / 2 for member in cluster])
        else:
            std_span = None
            std_pos = None
        score = calculate_score(cluster, std_span, std_pos, average_end - average_start, cluster[0].type)
        consolidated_clusters.append(SignatureClusterUniLocal(cluster[0].get_source()[0], int(round(average_start)), int(round(average_end)), score, len(cluster), cluster, cluster[0].type, std_span, std_pos))
    return consolidated_clusters


def consolidate_clusters_bilocal(clusters):
    """Consolidate clusters to a list of (type, contig, mean start, mean end, cluster size, members) tuples."""
    consolidated_clusters = []
    for cluster in clusters:
        #Source
        source_average_start = sum([member.get_source()[1] for member in cluster]) / len(cluster)
        source_average_end = sum([member.get_source()[2] for member in cluster]) / len(cluster)
        if len(cluster) > 1:
            source_std_span = stdev([member.get_source()[2] - member.get_source()[1] for member in cluster])
            source_std_pos = stdev([(member.get_source()[2] + member.get_source()[1]) / 2 for member in cluster])
        else:
            source_std_span = None
            source_std_pos = None

        if cluster[0].type == "dup":
            max_copies = max([member.copies for member in cluster])
            score = calculate_score(cluster, source_std_span, source_std_pos, source_average_end - source_average_start, cluster[0].type)
            consolidated_clusters.append(SignatureClusterBiLocal(cluster[0].get_source()[0],
                                                                int(round(source_average_start)),
                                                                int(round(source_average_end)),
                                                                cluster[0].get_source()[0],
                                                                int(round(source_average_end)),
                                                                int(round(source_average_end)) + max_copies *
                                                                (int(round(source_average_end)) -
                                                                 int(round(source_average_start))),
                                                                score, len(cluster), cluster, cluster[0].type, source_std_span, source_std_pos))
        else:
            #Destination
            destination_average_start = sum([member.get_destination()[1] for member in cluster]) / len(cluster)
            destination_average_end = sum([member.get_destination()[2] for member in cluster]) / len(cluster)
            if len(cluster) > 1:
                destination_std_span = stdev([member.get_destination()[2] - member.get_destination()[1] for member in cluster])
                destination_std_pos = stdev([(member.get_destination()[2] + member.get_destination()[1]) / 2 for member in cluster])
            else:
                destination_std_span = None
                destination_std_pos = None
            if source_std_span == None or source_std_pos == None or destination_std_span == None or destination_std_pos == None:
                score = calculate_score(cluster, None, None, mean([source_average_end - source_average_start, destination_average_end - destination_average_start]), cluster[0].type)
                consolidated_clusters.append(SignatureClusterBiLocal(cluster[0].get_source()[0], int(round(source_average_start)), int(round(source_average_end)),
                                                                cluster[0].get_destination()[0], int(round(destination_average_start)), int(round(destination_average_end)), score, len(cluster), cluster, cluster[0].type, None, None))
            else:
                score = calculate_score(cluster, mean([source_std_span, destination_std_span]), mean([source_std_pos, destination_std_pos]), mean([source_average_end - source_average_start, destination_average_end - destination_average_start]), cluster[0].type)
                consolidated_clusters.append(SignatureClusterBiLocal(cluster[0].get_source()[0], int(round(source_average_start)), int(round(source_average_end)),
                                                                cluster[0].get_destination()[0], int(round(destination_average_start)), int(round(destination_average_end)), score, len(cluster), cluster, cluster[0].type, mean([source_std_span, destination_std_span]), mean([source_std_pos, destination_std_pos])))
    return consolidated_clusters


def partition_and_cluster_candidates(candidates, options, type):
    partitions = form_partitions(candidates, options.distance_normalizer * options.cluster_max_distance)
    clusters = clusters_from_partitions(partitions, options)
    logging.info("Clustered {0}: {1} partitions and {2} clusters".format(type, len(partitions), len(clusters)))

    final_candidates = []
    for cluster in clusters:
        combined_score = max([candidate.score for candidate in cluster])
        combined_members = [member for candidate in cluster for member in candidate.members]
        try:
            combined_std_span = mean([candidate.std_span for candidate in cluster])
        except TypeError:
            combined_std_span = None
        try:
            combined_std_pos = mean([candidate.std_pos for candidate in cluster])
        except TypeError:
            combined_std_pos = None

        #Source
        source_average_start = (sum([candidate.get_source()[1] for candidate in cluster]) / len(cluster))
        source_average_end = (sum([candidate.get_source()[2] for candidate in cluster]) / len(cluster))

        #Destination
        destination_average_start = (sum([candidate.get_destination()[1] for candidate in cluster]) / len(cluster))
        destination_average_end = (sum([candidate.get_destination()[2] for candidate in cluster]) / len(cluster))

        #Origin deleted?
        cutpaste = False
        for member in cluster:
            if member.cutpaste:
                cutpaste = True

        if cluster[0].type == "dup_int":
            final_candidates.append(CandidateDuplicationInterspersed(cluster[0].get_source()[0], int(round(source_average_start)), int(round(source_average_end)),
                                                       cluster[0].get_destination()[0], int(round(destination_average_start)), int(round(destination_average_end)), combined_members, combined_score, combined_std_span, combined_std_pos, cutpaste))
    return final_candidates


def partition_and_cluster(signatures, options, type):
    partitions = form_partitions(signatures, options.distance_normalizer * options.cluster_max_distance)
    clusters = clusters_from_partitions(partitions, options)
    logging.info("Clustered {0}: {1} partitions and {2} clusters".format(type, len(partitions), len(clusters)))
    if type == "deleted regions" or type == "inserted regions" or type == "inverted regions":
        return sorted(consolidate_clusters_unilocal(clusters), key=lambda cluster: (cluster.contig, (cluster.end + cluster.start) / 2))
    elif type == "tandem duplicated regions" or type == "inserted regions with detected region of origin":
        return consolidate_clusters_bilocal(clusters)
    else:
        logging.error("Unknown parameter type={0} to function partition_and_cluster.")


import logging

class Signature:
    """Signature class for basic signatures of structural variants. An signature is always detected from a single read.
    """
    def __init__(self, contig, start, end, signature, read):
        self.contig = contig
        self.start = start
        self.end = end
        self.signature = signature
        self.read = read
        self.type = "unk"
        if self.end < self.start:
            logging.warning("Signature with invalid coordinates (end < start): " + self.as_string())


    def get_source(self):
        return (self.contig, self.start, self.end)


    def get_key(self):
        contig, start, end = self.get_source()
        return (self.type, contig, (start + end) // 2)


    def position_distance_to(self, signature2):
        """Return position distance between two signatures."""
        this_contig, this_start, this_end = self.get_source()
        other_contig, other_start, other_end = signature2.get_source()
        this_center = (this_start + this_end) // 2
        other_center = (other_start + other_end) // 2
        if self.type == signature2.type and this_contig == other_contig:
            return min(abs(this_start - other_start), abs(this_end - other_end), abs(this_center - other_center))
        else:
            return float("inf")


    def as_string(self, sep="\t"):
        contig, start, end = self.get_source()
        return sep.join(["{0}","{1}","{2}","{3}","{4}"]).format(contig, start, end, "{0};{1}".format(self.type, self.signature), self.read)


class SignatureDeletion(Signature):
    """SV Signature: a region (contig:start-end) has been deleted and is not present in sample"""
    def __init__(self, contig, start, end, signature, read):
        self.contig = contig
        #0-based start of the deletion (first deleted base)
        self.start = start
        #0-based end of the deletion (one past the last deleted base)
        self.end = end
        self.signature = signature
        self.read = read
        self.type = "del"


class SignatureInsertion(Signature):
    """SV Signature: a region of length end-start has been inserted at contig:start"""
    def __init__(self, contig, start, end, signature, read):
        self.contig = contig
        #0-based start of the insertion (base after the insertion)
        self.start = start
        #0-based start of the insertion (base after the insertion) + length of the insertion
        self.end = end
        self.signature = signature
        self.read = read
        self.type = "ins"


class SignatureInversion(Signature):
    """SV Signature: a region (contig:start-end) has been inverted in the sample"""
    def __init__(self, contig, start, end, signature, read, direction):
        self.contig = contig
        #0-based start of the inversion (first inverted base)
        self.start = start
        #0-based end of the inversion (one past the last inverted base)
        self.end = end
        self.signature = signature
        self.read = read
        self.type = "inv"
        self.direction = direction


    def as_string(self, sep="\t"):
        contig, start, end = self.get_source()
        return sep.join(["{0}","{1}","{2}","{3}","{4}"]).format(contig, start, end, "{0};{1};{2}".format(self.type, self.direction, self.signature), self.read)


class SignatureInsertionFrom(Signature):
    """SV Signature: a region (contig:start-end) has been inserted at contig2:pos in the sample"""
    def __init__(self, contig1, start, end, contig2, pos, signature, read):
        self.contig1 = contig1
        #0-based start of the region (first copied base)
        self.start = start
        #0-based end of the region (one past the last copied base)
        self.end = end

        self.contig2 = contig2
        #0-based start of the insertion (base after the insertion)
        self.pos = pos

        self.signature = signature
        self.read = read
        self.type = "ins_dup"


    def get_source(self):
        return (self.contig1, self.start, self.end)


    def get_destination(self):
        source_contig, source_start, source_end = self.get_source()
        return (self.contig2, self.pos, self.pos + (source_end - source_start))


    def get_key(self):
        source_contig, source_start, source_end = self.get_source()
        dest_contig, dest_start, dest_end = self.get_destination()
        return (self.type, source_contig, dest_contig, dest_start + (source_start + source_end) // 2)


    def as_string(self, sep="\t"):
        source_contig, source_start, source_end = self.get_source()
        dest_contig, dest_start, dest_end = self.get_destination()
        return sep.join(["{0}:{1}-{2}","{3}:{4}-{5}","{6}", "{7}"]).format(source_contig, source_start, source_end,
                                                                           dest_contig, dest_start, dest_end,
                                                                           "{0};{1}".format(self.type, self.signature), self.read)


class SignatureDuplicationTandem(Signature):
    """SV Signature: a region (contig:start-end) has been tandemly duplicated"""

    def __init__(self, contig, start, end, copies, signature, read):
        self.contig = contig
        #0-based start of the region (first copied base)
        self.start = start
        #0-based end of the region (one past the last copied base)
        self.end = end

        self.copies = copies

        self.signature = signature
        self.read = read
        self.type = "dup"


    def get_destination(self):
        source_contig, source_start, source_end = self.get_source()
        return (source_contig, source_end, source_end + self.copies * (source_end - source_start))


    def as_string(self, sep="\t"):
        source_contig, source_start, source_end = self.get_source()
        dest_contig, dest_start, dest_end = self.get_destination()
        return sep.join(["{0}:{1}-{2}","{3}:{4}-{5}","{6}", "{7}"]).format(source_contig, source_start, source_end,
                                                                           dest_contig, dest_start, dest_end,
                                                                           "{0};{1};{2}".format(self.type, self.signature, self.copies), self.read)


class SignatureTranslocation(Signature):
    """SV Signature: two positions (contig1:pos1 and contig2:pos2) are connected in the sample"""
    def __init__(self, contig1, pos1, direction1, contig2, pos2, direction2, signature, read):
        self.contig1 = contig1
        #0-based source of the translocation (first base before the translocation)
        self.pos1 = pos1
        self.direction1 = direction1
        self.contig2 = contig2
        #0-based destination of the translocation (first base after the translocation)
        self.pos2 = pos2
        self.direction2 = direction2
        self.signature = signature
        self.read = read
        self.type = "tra"


    def get_source(self):
        return (self.contig1, self.pos1, self.pos1 + 1)


    def get_destination(self):
        return (self.contig2, self.pos2, self.pos2 + 1)


    def as_string(self, sep="\t"):
        source_contig, source_start, source_end = self.get_source()
        dest_contig, dest_start, dest_end = self.get_destination()
        return sep.join(["{0}:{1}-{2}","{3}:{4}-{5}","{6}", "{7}"]).format(source_contig, source_start, source_end,
                                                                           dest_contig, dest_start, dest_end,
                                                                           "{0};{1}".format(self.type, self.signature), self.read)


    def get_key(self):
        return (self.type, self.contig1, self.pos1)


class SignatureClusterUniLocal(Signature):
    """Signature cluster class for clusters of signatures with only one genomic location.
    """
    def __init__(self, contig, start, end, score, size, members, type, std_span, std_pos):
        self.contig = contig
        self.start = start
        self.end = end
        
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.size = size
        self.members = members
        self.type = type


    def get_bed_entry(self):
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}".format(self.contig, self.start, self.end, "{0};{1};{2};{3}".format(self.type, self.size, self.std_span, self.std_pos), self.score, "["+"][".join([ev.as_string("|") for ev in self.members])+"]")


    def get_vcf_entry(self):
        if self.type == "del":
            svtype = "DEL"
        elif self.type == "ins":
            svtype = "INS"
        elif self.type == "inv":
            svtype = "INV"
        else:
            return
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}".format(self.contig, self.start+1, ".", "N", "<" + svtype + ">", ".", "PASS", "SVTYPE={0};END={1};SVLEN={2};STD_SPAN={3};STD_POS={4}".format(svtype, self.end, self.end - self.start, self.std_span, self.std_pos))


    def get_length(self):
        return self.end - self.start

class SignatureClusterBiLocal(Signature):
    """Signature cluster class for clusters of signatures with two genomic locations (source and destination).
    """
    def __init__(self, source_contig, source_start, source_end, dest_contig, dest_start, dest_end, score, size, members, type, std_span, std_pos):
        self.source_contig = source_contig
        self.source_start = source_start
        self.source_end = source_end
        self.dest_contig = dest_contig
        self.dest_start = dest_start
        self.dest_end = dest_end
        
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.size = size
        self.members = members
        self.type = type


    def get_source(self):
        return (self.source_contig, self.source_start, self.source_end)


    def get_destination(self):
        return (self.dest_contig, self.dest_start, self.dest_end)


    def get_bed_entries(self):
        source_entry = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}".format(self.source_contig, self.source_start, self.source_end, "{0}_source;{1}:{2}-{3};{4};{5};{6}".format(self.type, self.dest_contig, self.dest_start, self.dest_end, self.size, self.std_span, self.std_pos), self.score, "["+"][".join([ev.as_string("|") for ev in self.members])+"]")
        dest_entry = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}".format(self.dest_contig, self.dest_start, self.dest_end, "{0}_dest;{1}:{2}-{3};{4}".format(self.type, self.source_contig, self.source_start, self.source_end, self.size), self.score, "["+"][".join([ev.as_string("|") for ev in self.members])+"]")
        return (source_entry, dest_entry)


    def get_vcf_entry(self):
        if self.type == "dup":
            svtype = "DUP:TANDEM"
        else:
            return
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}".format(self.source_contig, self.source_start+1, ".", "N", "<" + svtype + ">", ".", "PASS", "SVTYPE={0};END={1};SVLEN={2};STD_SPAN={3};STD_POS={4}".format(svtype, self.source_end, self.source_end - self.source_start, self.std_span, self.std_pos))


    def get_source_length(self):
        return self.source_end - self.source_start


    def get_destination_length(self):
        return self.dest_end - self.dest_start


import os
import logging

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from svim.SVIM_clustering import partition_and_cluster
from svim.SVSignature import SignatureTranslocation


def complete_translocations(translocation_signatures):
    """Generate a complete list of translocation by adding all reversed translocations"""

    reversed_translocations = []
    for signature in translocation_signatures:
        reversed_translocations.append(SignatureTranslocation(signature.contig2, signature.pos2, 'fwd' if signature.direction2 == 'rev' else 'rev', signature.contig1, signature.pos1, 'fwd' if signature.direction1 == 'rev' else 'rev', signature.signature, signature.read))
    return translocation_signatures + reversed_translocations


def cluster_sv_signatures(sv_signatures, options):
    """Takes a list of SVSignatures and splits them up by type. The SVSignatures of each type are clustered and returned as a tuple of
    (deletion_signature_clusters, insertion_signature_clusters, inversion_signature_clusters, tandem_duplication_signature_clusters, insertion_from_signature_clusters, completed_translocation_signatures)."""

    deletion_signatures = [ev for ev in sv_signatures if ev.type == 'del']
    insertion_signatures = [ev for ev in sv_signatures if ev.type == 'ins']
    inversion_signatures = [ev for ev in sv_signatures if ev.type == 'inv']
    tandem_duplication_signatures = [ev for ev in sv_signatures if ev.type == 'dup']
    translocation_signatures = [ev for ev in sv_signatures if ev.type == 'tra']
    insertion_from_signatures = [ev for ev in sv_signatures if ev.type == 'ins_dup']

    # Cluster SV signatures
    deletion_signature_clusters = partition_and_cluster(deletion_signatures, options, "deleted regions")
    insertion_signature_clusters = partition_and_cluster(insertion_signatures, options, "inserted regions")
    inversion_signature_clusters = partition_and_cluster(inversion_signatures, options, "inverted regions")
    tandem_duplication_signature_clusters = partition_and_cluster(tandem_duplication_signatures, options, "tandem duplicated regions")
    insertion_from_signature_clusters = partition_and_cluster(insertion_from_signatures, options, "inserted regions with detected region of origin")

    return (deletion_signature_clusters, insertion_signature_clusters, inversion_signature_clusters, tandem_duplication_signature_clusters, insertion_from_signature_clusters, complete_translocations(translocation_signatures))


def write_signature_clusters_bed(working_dir, clusters):
    """Write signature clusters into working directory in BED format."""
    deletion_signature_clusters, insertion_signature_clusters, inversion_signature_clusters, tandem_duplication_signature_clusters, insertion_from_signature_clusters, completed_translocations = clusters

    # Print SV signature clusters
    if not os.path.exists(working_dir + '/signatures'):
        os.mkdir(working_dir + '/signatures')
    deletion_signature_output = open(working_dir + '/signatures/del.bed', 'w')
    insertion_signature_output = open(working_dir + '/signatures/ins.bed', 'w')
    inversion_signature_output = open(working_dir + '/signatures/inv.bed', 'w')
    tandem_duplication_signature_source_output = open(working_dir + '/signatures/dup_tan_source.bed', 'w')
    tandem_duplication_signature_dest_output = open(working_dir + '/signatures/dup_tan_dest.bed', 'w')
    translocation_signature_output = open(working_dir + '/signatures/trans.bed', 'w')
    insertion_from_signature_output = open(working_dir + '/signatures/ins_dup.bed', 'w')

    for cluster in deletion_signature_clusters:
        print(cluster.get_bed_entry(), file=deletion_signature_output)
    for cluster in insertion_signature_clusters:
        print(cluster.get_bed_entry(), file=insertion_signature_output)
    for cluster in inversion_signature_clusters:
        print(cluster.get_bed_entry(), file=inversion_signature_output)
    for cluster in tandem_duplication_signature_clusters:
        bed_entries = cluster.get_bed_entries()
        print(bed_entries[0], file=tandem_duplication_signature_source_output)
        print(bed_entries[1], file=tandem_duplication_signature_dest_output)
    for cluster in insertion_from_signature_clusters:
        bed_entries = cluster.get_bed_entries()
        print(bed_entries[0], file=insertion_from_signature_output)
        print(bed_entries[1], file=insertion_from_signature_output)
    for translocation in completed_translocations:
        print("{0}\t{1}\t{2}\t{3}\t{4}\t{5}".format(translocation.contig1, translocation.pos1, translocation.pos1+1, ">{0}:{1}".format(translocation.contig2, translocation.pos2), translocation.signature, translocation.read), file=translocation_signature_output)

    deletion_signature_output.close()
    insertion_signature_output.close()
    inversion_signature_output.close()
    tandem_duplication_signature_source_output.close()
    tandem_duplication_signature_dest_output.close()
    translocation_signature_output.close()
    insertion_from_signature_output.close()


def write_signature_clusters_vcf(working_dir, clusters, version):
    """Write signature clusters into working directory in VCF format."""
    deletion_signature_clusters, insertion_signature_clusters, inversion_signature_clusters, tandem_duplication_signature_clusters, insertion_from_signature_clusters, completed_translocations = clusters

    if not os.path.exists(working_dir + '/signatures'):
        os.mkdir(working_dir + '/signatures')
    vcf_output = open(working_dir + '/signatures/all.vcf', 'w')

    # Write header lines
    print("##fileformat=VCFv4.3", file=vcf_output)
    print("##source=SVIMV{0}".format(version), file=vcf_output)
    print("##ALT=<ID=DEL,Description=\"Deletion\">", file=vcf_output)
    print("##ALT=<ID=INV,Description=\"Inversion\">", file=vcf_output)
    print("##ALT=<ID=DUP,Description=\"Duplication\">", file=vcf_output)
    print("##ALT=<ID=DUP:TANDEM,Description=\"Tandem Duplication\">", file=vcf_output)
    print("##ALT=<ID=INS,Description=\"Insertion\">", file=vcf_output)
    print("##INFO=<ID=END,Number=1,Type=Integer,Description=\"End position of the variant described in this record\">", file=vcf_output)
    print("##INFO=<ID=SVTYPE,Number=1,Type=String,Description=\"Type of structural variant\">", file=vcf_output)
    print("##INFO=<ID=SVLEN,Number=.,Type=Integer,Description=\"Difference in length between REF and ALT alleles\">", file=vcf_output)
    print("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO", file=vcf_output)

    vcf_entries = []
    for cluster in deletion_signature_clusters:
        vcf_entries.append((cluster.get_source(), cluster.get_vcf_entry()))
    for cluster in insertion_signature_clusters:
        vcf_entries.append((cluster.get_source(), cluster.get_vcf_entry()))
    for cluster in inversion_signature_clusters:
        vcf_entries.append((cluster.get_source(), cluster.get_vcf_entry()))
    for cluster in tandem_duplication_signature_clusters:
        vcf_entries.append((cluster.get_source(), cluster.get_vcf_entry()))

    # Sort and write entries to VCF
    for source, entry in sorted(vcf_entries, key=lambda pair: pair[0]):
        print(entry, file=vcf_output)

    vcf_output.close()


def plot_histograms(working_dir, clusters):
    deletion_signature_clusters, insertion_signature_clusters, inversion_signature_clusters, tandem_duplication_signature_clusters, insertion_from_signature_clusters, completed_translocations = clusters

    if not os.path.exists(working_dir + '/signatures'):
        os.mkdir(working_dir + '/signatures')
    pdf = PdfPages(working_dir + '/signatures/signature_cluster_histograms.pdf')
    fig = plt.figure()
    fig.suptitle('Deleted region signature clusters', fontsize=10)

    plt.subplot(2, 1, 1)
    deletion_scores = [del_cluster.score for del_cluster in deletion_signature_clusters]
    plt.hist(deletion_scores, bins=100)
    plt.xlabel('Score')
    plt.ylabel('Count')
    plt.yscale('log', nonposy='clip')
    plt.title('Histogram of Score')
    plt.grid(True)

    plt.subplot(2, 1, 2)
    deletion_sizes = [del_cluster.end - del_cluster.start for del_cluster in deletion_signature_clusters]
    plt.hist(deletion_sizes, bins=20)
    plt.xlabel('Size in bp')
    plt.ylabel('Count')
    plt.yscale('log', nonposy='clip')
    plt.title('Histogram of Size')
    plt.grid(True)

    pdf.savefig(fig)

    fig = plt.figure()
    fig.suptitle('Inserted regions', fontsize=10)

    plt.subplot(2, 1, 1)
    insertion_scores = [ins_cluster.score for ins_cluster in insertion_signature_clusters]
    plt.hist(insertion_scores, bins=100)
    plt.xlabel('Score')
    plt.ylabel('Count')
    plt.yscale('log', nonposy='clip')
    plt.title('Histogram of Score')
    plt.grid(True)

    plt.subplot(2, 1, 2)
    insertion_sizes = [ins_cluster.end - ins_cluster.start for ins_cluster in insertion_signature_clusters]
    plt.hist(insertion_sizes, bins=20)
    plt.xlabel('Size in bp')
    plt.ylabel('Count')
    plt.yscale('log', nonposy='clip')
    plt.title('Histogram of Size')
    plt.grid(True)

    pdf.savefig(fig)

    fig = plt.figure()
    fig.suptitle('Inverted regions', fontsize=10)

    plt.subplot(2, 1, 1)
    inversion_scores = [inv_cluster.score for inv_cluster in inversion_signature_clusters]
    plt.hist(inversion_scores, bins=100)
    plt.xlabel('Score')
    plt.ylabel('Count')
    plt.yscale('log', nonposy='clip')
    plt.title('Histogram of Score')
    plt.grid(True)

    plt.subplot(2, 1, 2)
    inversion_sizes = [inv_cluster.end - inv_cluster.start for inv_cluster in inversion_signature_clusters]
    plt.hist(inversion_sizes, bins=20)
    plt.xlabel('Size in bp')
    plt.ylabel('Count')
    plt.yscale('log', nonposy='clip')
    plt.title('Histogram of Size')
    plt.grid(True)

    pdf.savefig(fig)

    pdf.close()



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the relevant file
with open(os.path.join(here, 'README.rst')) as f:
      long_description = f.read()

def get_version(string):
      """ Parse the version number variable __version__ from a script. """
      import re
      version_re = r"^__version__ = ['\"]([^'\"]*)['\"]"
      version_str = re.search(version_re, string, re.M).group(1)
      return version_str


setup(name='svim',
      version=get_version(open('src/svim/svim').read()),
      description='A structural variant caller for long reads.',
      long_description=long_description,
      url='https://github.com/eldariont/svim',
      author='David Heller',
      author_email='heller_d@molgen.mpg.de',
      license='GPLv3',
      classifiers=[
      'Development Status :: 4 - Beta',
      'Environment :: Console',
      'Intended Audience :: Science/Research',
      'Topic :: Scientific/Engineering :: Bio-Informatics',
      'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
      'Programming Language :: Python :: 3.6'
      ],
      keywords='svim SV PacBio structural variation caller',
      packages = find_packages("src"),
      package_dir = {"": "src"},
      data_files = [("", ["LICENSE"])],
      zip_safe=False,
      install_requires=['pysam', 'numpy', 'scipy', 'biopython', 'matplotlib'],
      scripts=['src/svim/svim'])



import unittest

from svim.SVIM_inter import is_similar

class TestSVIMInter(unittest.TestCase):

    def test_is_similar(self):
        self.assertFalse(is_similar("chrI", 0, 100, "chrII", 0, 100))
        self.assertTrue(is_similar("chrI", 0, 100, "chrI", 0, 100))
        self.assertTrue(is_similar("chrI", 0, 100, "chrI", 10, 90))
        self.assertFalse(is_similar("chrI", 0, 100, "chrI", 21, 100))

if __name__ == '__main__':
    unittest.main()



import sys
import os
import logging
import argparse


def parse_arguments(program_version, arguments = sys.argv[1:]):
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="""SVIM (pronounced SWIM) is a structural variant caller for long reads. 
It discriminates five different variant classes: deletions, tandem and interspersed duplications, 
inversions and novel element insertions. SVIM is unique in its capability of extracting both the genomic origin and 
destination of duplications.

SVIM consists of three major steps:
- COLLECT detects signatures for SVs in long read alignments
- CLUSTER merges signatures that come from the same SV
- COMBINE combines clusters from different genomic regions and classifies them into distinct SV types

SVIM can process two types of input. Firstly, it can detect SVs from raw reads by aligning them to a given reference genome first ("SVIM.py reads [options] working_dir reads genome").
Alternatively, it can detect SVs from existing reads alignments in SAM/BAM format ("SVIM.py alignment [options] working_dir bam_file").
""")
    subparsers = parser.add_subparsers(help='modes', dest='sub')
    parser.add_argument('--version', '-v', action='version', version='%(prog)s {version}'.format(version=program_version))

    parser_fasta = subparsers.add_parser('reads', help='Detect SVs from raw reads. Align reads to given reference genome first.')
    parser_fasta.add_argument('working_dir', type=str, help='working directory')
    parser_fasta.add_argument('reads', type=str, help='Read file (FASTA, FASTQ, gzipped FASTA and FASTQ)')
    parser_fasta.add_argument('genome', type=str, help='Reference genome file (FASTA)')
    group_fasta_collect = parser_fasta.add_argument_group('COLLECT')
    group_fasta_collect.add_argument('--min_mapq', type=int, default=20, help='Minimum mapping quality of reads to consider')
    group_fasta_collect.add_argument('--min_sv_size', type=int, default=40, help='Minimum SV size to detect')
    group_fasta_collect.add_argument('--max_sv_size', type=int, default=100000, help='Maximum SV size to detect')
    group_fasta_collect.add_argument('--skip_indel', action='store_true', help='disable signature collection from within read alignments')
    group_fasta_collect.add_argument('--skip_segment', action='store_true', help='disable signature collection from between read alignments')
    group_fasta_collect.add_argument('--cores', type=int, default=1, help='CPU cores to use for alignment with ngmlr')
    group_fasta_collect.add_argument('--aligner', type=str, default="ngmlr", choices=["ngmlr", "minimap2"], help='tool for read alignment: ngmlr or minimap2 (default: ngmlr)')
    group_fasta_collect.add_argument('--nanopore', action='store_true', help='use Nanopore settings for read alignment (default: off)')
    group_fasta_collect.add_argument('--segment_gap_tolerance', type=int, default=10, help='Maximum tolerated gap between adjacent alignment segments')
    group_fasta_collect.add_argument('--segment_overlap_tolerance', type=int, default=5, help='Maximum tolerated overlap between adjacent alignment segments')
    group_fasta_cluster = parser_fasta.add_argument_group('CLUSTER')
    group_fasta_cluster.add_argument('--partition_max_distance', type=int, default=5000, help='Maximum distance in bp between SVs in a partition (default: 5000)')
    group_fasta_cluster.add_argument('--distance_normalizer', type=int, default=900, help='Distance normalizer used for span-position distance (default: 900)')
    group_fasta_cluster.add_argument('--cluster_max_distance', type=float, default=0.3, help='Maximum span-position distance between SVs in a cluster (default: 0.3)')
    group_fasta_combine = parser_fasta.add_argument_group('COMBINE')
    group_fasta_combine.add_argument('--del_ins_dup_max_distance', type=float, default=1.0, help='Maximum span-position distance between the origin of an insertion and a deletion to be flagged as a potential cut&paste insertion')
    group_fasta_combine.add_argument('--trans_destination_partition_max_distance', type=int, default=1000, help='Maximum distance in bp between translocation breakpoint destinations in a partition')
    group_fasta_combine.add_argument('--trans_partition_max_distance', type=int, default=200, help='Maximum distance in bp between translocation breakpoints in a partition')
    group_fasta_combine.add_argument('--trans_sv_max_distance', type=int, default=500, help='Maximum distance in bp between a translocation breakpoint and an SV signature to be combined')
    group_fasta_combine.add_argument('--sample', type=str, default="Sample", help='Sample ID to include in output vcf (default: Sample)')

    parser_bam = subparsers.add_parser('alignment', help='Detect SVs from an existing alignment')
    parser_bam.add_argument('working_dir', type=os.path.abspath, help='working directory')
    parser_bam.add_argument('bam_file', type=str, help='SAM/BAM file with aligned long reads (sorted either by coordinate or queryname)')
    group_bam_collect = parser_bam.add_argument_group('COLLECT')
    group_bam_collect.add_argument('--min_mapq', type=int, default=20, help='Minimum mapping quality of reads to consider')
    group_bam_collect.add_argument('--min_sv_size', type=int, default=40, help='Minimum SV size to detect')
    group_bam_collect.add_argument('--max_sv_size', type=int, default=100000, help='Maximum SV size to detect')
    group_bam_collect.add_argument('--skip_indel', action='store_true', help='disable signature collection from within read alignments')
    group_bam_collect.add_argument('--skip_segment', action='store_true', help='disable signature collection from between read alignments')
    group_bam_collect.add_argument('--segment_gap_tolerance', type=int, default=10, help='Maximum tolerated gap between adjacent alignment segments')
    group_bam_collect.add_argument('--segment_overlap_tolerance', type=int, default=5, help='Maximum tolerated overlap between adjacent alignment segments')
    group_bam_cluster = parser_bam.add_argument_group('CLUSTER')
    group_bam_cluster.add_argument('--partition_max_distance', type=int, default=5000, help='Maximum distance in bp between SVs in a partition (default: 5000)')
    group_bam_cluster.add_argument('--distance_normalizer', type=int, default=900, help='Distance normalizer used for span-position distance (default: 900)')
    group_bam_cluster.add_argument('--cluster_max_distance', type=float, default=0.3, help='Maximum span-position distance between SVs in a cluster (default: 0.3)')
    group_bam_combine = parser_bam.add_argument_group('COMBINE')
    group_bam_combine.add_argument('--del_ins_dup_max_distance', type=float, default=1.0, help='Maximum span-position distance between the origin of an insertion and a deletion to be flagged as a potential cut&paste insertion')
    group_bam_combine.add_argument('--trans_destination_partition_max_distance', type=int, default=1000, help='Maximum distance in bp between translocation breakpoint destinations in a partition')
    group_bam_combine.add_argument('--trans_partition_max_distance', type=int, default=200, help='Maximum distance in bp between translocation breakpoints in a partition')
    group_bam_combine.add_argument('--trans_sv_max_distance', type=int, default=500, help='Maximum distance in bp between a translocation breakpoint and an SV signature to be combined')
    group_bam_combine.add_argument('--sample', type=str, default="Sample", help='Sample ID to include in output vcf (default: Sample)')

    return parser.parse_args(arguments)


def guess_file_type(reads_path):
    if reads_path.endswith(".fa") or reads_path.endswith(".fasta") or reads_path.endswith(".FA"):
        logging.info("Recognized reads file as FASTA format.")
        return "fasta"
    elif reads_path.endswith(".fq") or reads_path.endswith(".fastq") or reads_path.endswith(".FQ"):
        logging.info("Recognized reads file as FASTQ format.")
        return "fastq"
    elif reads_path.endswith(".fa.gz") or reads_path.endswith(".fasta.gz") or reads_path.endswith(".FA.gz") or reads_path.endswith(".fa.gzip") or reads_path.endswith(".fasta.gzip") or reads_path.endswith(".FA.gzip"):
        logging.info("Recognized reads file as gzipped FASTA format.")
        return "fasta_gzip"
    elif reads_path.endswith(".fq.gz") or reads_path.endswith(".fastq.gz") or reads_path.endswith(".FQ.gz") or reads_path.endswith(".fq.gzip") or reads_path.endswith(".fastq.gzip") or reads_path.endswith(".FQ.gzip"):
        logging.info("Recognized reads file as gzipped FASTQ format.")
        return "fastq_gzip"
    elif reads_path.endswith(".fa.fn") or reads_path.endswith(".fasta.fn") or reads_path.endswith(".FA.fn") or reads_path.endswith(".fq.fn") or reads_path.endswith(".fastq.fn") or reads_path.endswith(".FQ.fn"):
        logging.info("Recognized reads file as file list format.")
        return "list"
    else:
        logging.error("Unknown file ending of file {0}. See github.com/eldariont/svim/wiki/ for supported file endings. Exiting.".format(reads_path))
        return "unknown"


def read_file_list(path):
    file_list = open(path, "r")
    for line in file_list:
        yield line.strip()
    file_list.close()